def main(fn):
    import os
    from reportlab.lib.utils import open_and_read
    from reportlab.lib.pdfencrypt import encryptPdfInMemory
    pdf_bytes = open_and_read(fn)
    ofn = os.path.splitext(os.path.basename(fn))[0]
    with open(ofn+'-encrypted.pdf','wb') as f:
        f.write(encryptPdfInMemory(pdf_bytes, 'master', 'user', canModify=0, canCopy=0, strength=128))

if __name__=='__main__':
    import sys
    main(sys.argv[1])



__code__='GatU0D/[on&>$QoP(Q(,JX3cWGtE-$\\,uf2WXrol"H)7[d2`dE[\'C[t-8G&DAp5o=8Dt4QI-H.G?gGqiL\'4:r_;4O(]fS$r1UEYZU%lm]r0=/TQNUWJ+\\<T=!!U";GRO$(:k01kU;c1VW!XZ,.UW^i%bLbD"?6QE%G1eG%!VpE42@VA`rK`PNs!Z/ScG8_4:-6>n9_qm!)E]g!B^a7%e)U=O+?hTEpQekYo`gk%fWKDYQi!+W!5%S2,JF?>6XQn+\'<qoK9q6DRH9M9>_#-+1<!W)M:JhhY<qi)Nc0r2E\\UC$\']M1T_^b9`/_13.,\\k<bZMGe@LO_MgVqtCh!pn=5Z?IL#\\mR&M:=W4qX$CjCe0Cds4Sg:B\\]bY3jtKAV_>#PZa1\\BD%pBBQ&!>L@2R$6:3isJ.Yu,Dj\'YCI?<0(%^?ZRUb\\r/>K9RC;./0h5B/IpD>F+o:-e\\pb2qI9>FHgskOkce!%FiH=s9kV,)`n8R[SU$O+1Gd#4U+\']nN9KAj]]K#^0N,W\'S*jjPi]-(%<p6B:l;*bn?V@,#hNhTt&2(EK"p;2*:YAfPWi[)=aVF9t5Xn3I16+\\d(kj`o6mo>s?/MQhBl3Q:VmE_9e(/Cm%bB?B3bljnF3cXZH7j%>nMmaMR`JH$29@%:*lBhh4m7o7ag@A\'HA3[2VHT$:>8$b+<S,<?<aX9]2OA+I8`8>q)s1na<,+k+p4L2eB.b56W\'L+d+HM#kgi@3D)c5rd.^_H5fbK+&fiX$!JPkgiI(<mNddu%LRl>$9]m8f7h8Q"EUM"34!;"6f<B:HGSESCZY5*nid>6+5!])j?aCJ[hZ-6Ks4Ysue\\uEQK.+p")Sf`8O%^3O!ndZI0NotO,]&Na/9C,\'L#>qS,E:Ce&pF2nB`OGTsq&QDQX<Y>-T1<V^\'\'kto(LXrC(=u)N(:a9L$#3aYi/MLT<*4E9TQ[Qb*T;,S<aA@+lX#KXgim,T(e14lrB;;a%9MYTY:f,F43_l3f^ioaF0rKZ7>[MGSglgamg!i+"JBQg9ZZn\'aJtABFLI^pmQ9k.hcf\\*n9]FB\\^-;\\<EK,H)Ui/!2#cZf\\hYWOQ2^h)l@*g~>'



'''
functions for solving the internal rate of return problem
'''
__all__=('npv','dnpv','irr')
from rlextra.utils.newton import solve as newton_solve
def npv(r,C):
    '''
    compute the net present value of the cash flows given in C assuming rate r


    >>> print('%.5f' % npv(0.1, [-100.0, 60.0, 60.0, 60.0]))
    49.21112
    '''
    r += 1.0
    return sum(c/r**t for t,c in enumerate(C))

def dnpv(r,C):
    '''
    compute d npv(r,C)/dr

    >>> dnpv(0.1, [-100.0, 60.0, 60.0, 60.0])
    -262.68697493340613
    >>> abs(dnpv(0.1, [-100.0, 60.0, 60.0, 60.0])-(npv(0.1+0.5e-8, [-100.0, 60.0, 60.0, 60.0])-npv(0.1-0.5e-8, [-100.0, 60.0, 60.0, 60.0]))/1e-8)<1e-5  #test the drivative against finite difference
    True
    '''
    r += 1.0
    return sum(-t*c/r**(t+1) for t,c in enumerate(C))

def irr(C, r=0.001, maxIterations=100, precision=1e-8, failHard=True):
    """The IRR or Internal Rate of Return is the annualized effective 
       compounded return rate which can be earned on the invested 
       capital, i.e., the yield on the investment.
       
       >>> print('%.8f' % irr([-100.0, 60.0, 60.0, 60.0]))
       0.36309654

    This function assumes unit time. To convert to other periods use the appropriate calculation.
    If C represents monthly cash flows then the result will be a monthly rate, rm. 
    The annual rate is then ry=(1+rm)*12 - 1
    """
    return newton_solve(
            lambda r,C=C: npv(r,C),         #f(r)
            lambda r,C=C: dnpv(r,C),        #f'(r)
            0,                              #we want func(r)=0
            x=r,                            #initial guess
            maxIterations=maxIterations,
            precision = precision,
            failHard = failHard,
            )

def _doctest():
    import doctest
    doctest.testmod()

if __name__ == "__main__":
    _doctest()



__code__='Gat=(s)6@U\'RuIZTe-[Fl.FWrG^S2]UcC[@:"(qYKlf\\&Z\\fGPKh[m!GAU%pgRR`d<%:*IBts#<g0_;_c8J7L1o.SM48\\I&kKC4#754=WSZ\\UK.fQ,l2dljY$OT.L\\%qOD!@%VC^bP?\'j=cT(5l*>=*Pa\\N(>p(GH,\'U!7DQ6[k5[^J4:cZ;E!c"Xi%j4:^_OYG!A=P(!r+5&\'u8,q56nh.GRa-M_!ApZiDl)PT)<jKrZH(`7WT\\VX?\\fhkMWiViq;1_I-JF>@WqjfA>pg%E!$=3>@bVW$H#D6(3s`EoeY@88`9rJ-`nCE3%DTI2YC*)+[$=hc4m0ZiUG>:?!YJ?\\Qf2r"C-E)"DpGcRLDLaMdJb/lC"N7X;s985lS:M._)e)!^peH>S?bT=c1&_,1q/6[(k:@ZF8Fu7WX=$NmscRb4*>QQk[uo(=at6D^m"!Ej"D$QPLRQ-eYq7;Suh\\`K"ZZ\\<XMUQ4;sd"dR/c<oLNgh?.]h@Acq!Vjclf*@F:rCh3p!q4UO/Es=/WD2B(JZ<.a*D<U),^QcuY6\'-h2)$FR!5)_9S#sY#fO0;9F<9psm=Xkrk2@H[NOt7fp6#^SPp-g`S_1@9FoZ)T8=f9Yq-M\'oF5MU9<r9"/@0&4fJi^fr#6oS2C37;]Q4gT\'.lhpWTkMY:6;g\'o&\\%F+.HZ"\\qpPh[QDa-ohW8fFO1ZcbN*\\0R9ZMa2;*7Ifc;l9DG/TrqH9Wi.S+3a!gG:)WC]@RK6K`;nHQ`=Gq3m1bSM,qAa=SL("PD1QApBXCk)s4k7PNn6E2r2u4Y@Mj5cHV?+G/HXmAbp`1S>YoVh0XA1pZUJO#EMp#X-T?HjV,e/VHU!aA6GV>hn/#$\\r"J?4saf\\IT=LoE3%_+ZKmu+)P^RUmD\\<b?1Ds_modC$/_\'7YG#@i+h)2MoZYi(am4_9\\<WKnfO1=&.m8`\'#$]@*.O\\rAIFC2sh@LF.*SXI7G<o%ZgCRTXqc/Y(n^$uP2/On@JL\\jYqHK-1XD+!:[r_+E7:*[c@VUP7A:!E<so#u*@SGK1hjM2sHA)WHU676Y3hk/=*g[fQ7$\\C\'(86DjVdo`U")/D5bm)#?Rq^4G^\'R6YR1r<D;%qU+aO]dQ[p]ZIHXTbsP>6"2DV6M4JNILslaeH$~>'



# topsort takes a list of pairs, where each pair (x, y) is taken to
# mean that x <= y wrt some abstract partial ordering.  The return
# value is a list, representing a total ordering that respects all
# the input constraints.
# E.g.,
#    topsort( [(1,2), (3,3)] )
# may return any of (but nothing other than)
#    [3, 1, 2]
#    [1, 3, 2]
#    [1, 2, 3]
# because those are the permutations of the input elements that
# respect the "1 precedes 2" and "3 precedes 3" input constraints.
# Note that a constraint of the form (x, x) is really just a trick
# to make sure x appears *somewhere* in the output list.
#
# If there's a cycle in the constraints, say
#    topsort( [(1,2), (2,1)] )
# then CycleError is raised, and the exception object supports
# many methods to help analyze and break the cycles.  This requires
# a good deal more code than topsort itself!
# Tim Peters http://www.faqts.com/knowledge_base/view.phtml/aid/4294

class CycleError(ValueError):
    def __init__(self, sofar, numpreds, succs):
        ValueError.__init__(self, "cycle in constraints",
                           sofar, numpreds, succs)
        self.preds = None

    # return as much of the total ordering as topsort was able to
    # find before it hit a cycle
    def get_partial(self):
        return self[1]

    # return remaining elt -> count of predecessors map
    def get_pred_counts(self):
        return self[2]

    # return remaining elt -> list of successors map
    def get_succs(self):
        return self[3]

    # return remaining elements (== those that don't appear in
    # get_partial())
    def get_elements(self):
        return list(self.get_pred_counts().keys())

    # Return a list of pairs representing the full state of what's
    # remaining (if you pass this list back to topsort, it will raise
    # CycleError again, and if you invoke get_pairlist on *that*
    # exception object, the result will be isomorphic to *this*
    # invocation of get_pairlist).
    # The idea is that you can use pick_a_cycle to find a cycle,
    # through some means or another pick an (x,y) pair in the cycle
    # you no longer want to respect, then remove that pair from the
    # output of get_pairlist and try topsort again.
    def get_pairlist(self):
        succs = self.get_succs()
        answer = []
        for x in self.get_elements():
            if x in succs:
                for y in succs[x]:
                    answer.append( (x, y) )
            else:
                # make sure x appears in topsort's output!
                answer.append( (x, x) )
        return answer

    # return remaining elt -> list of predecessors map
    def get_preds(self):
        if self.preds is not None:
            return self.preds
        self.preds = preds = {}
        remaining_elts = self.get_elements()
        for x in remaining_elts:
            preds[x] = []
        succs = self.get_succs()

        for x in remaining_elts:
            if x in succs:
                for y in succs[x]:
                    preds[y].append(x)

        if __debug__:
            for x in remaining_elts:
                assert len(preds[x]) > 0
        return preds

    # return a cycle [x, ..., x] at random
    def pick_a_cycle(self):
        remaining_elts = self.get_elements()

        # We know that everything in remaining_elts has a predecessor,
        # but don't know that everything in it has a successor.  So
        # crawling forward over succs may hit a dead end.  Instead we
        # crawl backward over the preds until we hit a duplicate, then
        # reverse the path.
        preds = self.get_preds()
        from random import choice
        x = choice(remaining_elts)
        answer = []
        index = {}
        while not x in index:
            index[x] = len(answer) # index of x in answer
            answer.append(x)
            x = choice(preds[x])
        answer.append(x)
        answer = answer[index[x]:]
        answer.reverse()
        return answer

def topsort(pairlist):
    numpreds = {}   # elt -> # of predecessors
    successors = {} # elt -> list of successors
    for first, second in pairlist:
        # make sure every elt is a key in numpreds
        if first not in numpreds:
            numpreds[first] = 0
        if second not in numpreds:
            numpreds[second] = 0

        # if they're the same, there's no real dependence
        if first == second:
            continue

        # since first < second, second gains a pred ...
        numpreds[second] += 1

        # ... and first gains a succ
        if first in successors:
            successors[first].append(second)
        else:
            successors[first] = [second]

    # suck up everything without a predecessor
    answer = list(filter(lambda x, numpreds=numpreds: numpreds[x] == 0,
                    list(numpreds.keys())))

    # for everything in answer, knock down the pred count on
    # its successors; note that answer grows *in* the loop
    for x in answer:
        assert numpreds[x] == 0
        del numpreds[x]
        if x in successors:
            for y in successors[x]:
                numpreds[y] -= 1
                if numpreds[y] == 0:
                    answer.append(y)
            # following "del" isn't needed; just makes
            # CycleError details easier to grasp
            del successors[x]

    if numpreds:
        # everything in numpreds has at least one predecessor ->
        # there's a cycle
        if __debug__:
            for x in list(numpreds.keys()):
                assert numpreds[x] > 0
        raise CycleError(answer, numpreds, successors)
    return answer

def test():
    print(topsort( [(1,2), (3,4), (5,6), (1,3), (1,5), (1,6), (2,5)] ))

if __name__=='__main__':
    test()



#!/usr/bin/env python

"""Testsuite for svglib.

This tests basic functionality. Run with one of these lines from
inside the test directory:

    py.test -v -s test_basic.py
"""
import unittest
from reportlab.lib.testutils import makeSuiteForClasses
import io, logging, os
import textwrap, subprocess
from rlextra.thirdparty.svg import svglib
try:
    from lxml import etree
    def parseXML(s):
        return svglib.NodeCompat(etree.XML(s))
except ImportError:
    import xml.dom.minidom 
    def parseXML(s):
        return svglib.NodeCompat(xml.dom.minidom.parseString(s).documentElement)

from reportlab.graphics.shapes import (
    _CLOSEPATH, _CURVETO, _LINETO, _MOVETO, Group, Path, Polygon, PolyLine, Rect,
    )
from reportlab.lib import colors
from reportlab.lib.units import cm, inch
from reportlab.pdfgen.canvas import FILL_EVEN_ODD

svglib.logger.addHandler(logging.StreamHandler(open(os.path.join(os.path.dirname(__file__),'text_svg_basic.log'),'w')))

def _testit(func, mapping):
    "Call `func` on input in mapping and return list of failed tests."

    failed = []
    for input, expected in mapping:
        result = func(input)
        if not result == expected:
            failed.append((input, result, expected))

    if failed:
        print("failed tests (input, result, expected):")
        for input, result, expected in failed:
            print("  %s : %s != %s" % (repr(input), result, expected))

    return failed


class SVGBasicTest(unittest.TestCase):
    pass

class TestPaths(SVGBasicTest):
    """Testing path-related code."""

    def test_path_normalisation(self):
        "Test path normalisation."

        mapping = (
            ("",
                []),

            ("M10 20, L 30 40 ",
                ["M", [10, 20], "L", [30, 40]]),

            ("M10 20, L 40 40Z",
                ["M", [10, 20], "L", [40, 40], "Z", []]),

            ("M10 20, L 30 40 40 40Z",
                ["M", [10, 20], "L", [30, 40], "L", [40, 40], "Z", []]),

            ("  M10 20,L30 40,40 40Z  ",
                ["M", [10, 20], "L", [30, 40], "L", [40, 40], "Z", []]),

            ("  M 10 20, M 30 40, L 40 40, Z M 40 50, l 50 60, Z",
                ["M", [10, 20], "L", [30, 40], "L", [40, 40], "Z", [],
                 "M", [40, 50], "l", [50, 60], "Z", []]),

            ("  m 10 20, m 30 40, l 40 40, z",
                ["m", [10, 20], "l", [30, 40], "l", [40, 40], "z", []]),

            ("  m 10,20 30,40, l 40 40, z ",
                ["m", [10, 20], "l", [30, 40], "l", [40, 40], "z", []]),

            ("M 10,20 30,40, l 40 40, z",
                ["M", [10, 20], "L", [30, 40], "l", [40, 40], "z", []]),

            ("M0,0 500,300M500,0 0,300",
                ["M", [0, 0], "L", [500, 300], "M", [500, 0], "L", [0, 300]]),

            ("M10 20, l 5e-5,0",
                ["M", [10, 20], "l", [5e-5, 0]]),

            ("m246.026 120.178c-.558-.295-1.186-.768-1.395-1.054-.314-.438-.132-.456 1.163-.104 "
             "2.318.629 3.814.383 5.298-.873l1.308-1.103 1.54.784c.848.428 1.748.725 "
             "2.008.656.667-.176 2.05-1.95 2.005-2.564-.054-.759.587-.568.896.264.615 1.631-.281 "
             "3.502-1.865 3.918-.773.201-1.488.127-2.659-.281-1.438-.502-1.684-.494-2.405.058-1.618 "
             "1.239-3.869 1.355-5.894.299z",
                ['m', [246.026, 120.178], 'c', [-0.558, -0.295, -1.186, -0.768, -1.395, -1.054],
                 'c', [-0.314, -0.438, -0.132, -0.456, 1.163, -0.104],
                 'c', [2.318, 0.629, 3.814, 0.383, 5.298, -0.873],
                 'l', [1.308, -1.103], 'l', [1.54, 0.784],
                 'c', [0.848, 0.428, 1.748, 0.725, 2.008, 0.656],
                 'c', [0.667, -0.176, 2.05, -1.95, 2.005, -2.564],
                 'c', [-0.054, -0.759, 0.587, -0.568, 0.896, 0.264],
                 'c', [0.615, 1.631, -0.281, 3.502, -1.865, 3.918],
                 'c', [-0.773, 0.201, -1.488, 0.127, -2.659, -0.281],
                 'c', [-1.438, -0.502, -1.684, -0.494, -2.405, 0.058],
                 'c', [-1.618, 1.239, -3.869, 1.355, -5.894, 0.299],
                 'z', []
                ]),
        )
        failed = _testit(svglib.normalise_svg_path, mapping)
        self.assertEqual(len(failed) ,0)

    def test_relative_move_after_closepath(self):
        """
        A relative subpath is relative to the point *after* the previous
        closePath op (which is not recorded in path.points).
        """
        converter = svglib.Svg2RlgShapeConverter(None)
        node = svglib.NodeTracker(parseXML(
            '<path d="M0 0,0 1,1 1z m-1-1 0 1 1 0z"/>'
        ))
        # last point of this path should be 0 0
        path = converter.convertPath(node).contents[0]
        self.assertEqual(path.points[-2:] ,[0, 0])

    def test_cubic_bezier_shorthand(self):
        # If there is no previous command or if the previous command was not
        # an C, c, S or s, assume the first control point is coincident with
        # the current point.
        converter = svglib.Svg2RlgShapeConverter(None)
        node = svglib.NodeTracker(parseXML(
            '<path d="M3,4c-0.5-0.25-1.3-0.77-1.3-1.8h2.5s0.04,0.46,0.04,1.48z"/>'
        ))
        path = converter.convertPath(node).contents[0]
        self.assertEqual(path.operators ,[_MOVETO, _CURVETO, _LINETO, _CURVETO, _CLOSEPATH])
        self.assertEqual(path.points,
                [3.0, 4.0, 2.5, 3.75, 1.7, 3.23, 1.7, 2.2, 4.2, 2.2, 4.2, 2.2, 4.24, 2.66,
                4.24, 3.68,]
                )

    def test_elliptical_arc(self):
        converter = svglib.Svg2RlgShapeConverter(None)
        node = svglib.NodeTracker(parseXML(
                '<path d="M334.500000 0.000000 A185.000000 185.000000 0 0 1 334.500000 0.000000 '
                'L334.500000 185.000000 A0.000000 0.000000 0 0 0 334.500000 185.000000 z"/>'
                ))
        # First elliptical arc with identical start/end points, ignored
        path = converter.convertPath(node).contents[0]
        self.assertEqual(path.points, [334.5, 0.0, 334.5, 185.0, 334.5, 185.0])

    def test_unclosed_paths(self):
        converter = svglib.Svg2RlgShapeConverter(None)
        node = svglib.NodeTracker(parseXML('<path d="M0,0 4.5,3 0,6M4.5,3H9" id="W"/>'))
        group = converter.convertPath(node)
        self.assertEqual(len(group.contents) ,1)
        #closed_path = group.contents[0]
        #unclosed_path = group.contents[1]
        #self.assertEqual(len(closed_path.points) ,len(unclosed_path.points))
        #self.assertEqual(closed_path.operators ,[)
        #   _MOVETO, _LINETO, _LINETO, _CLOSEPATH, _MOVETO, _LINETO, _CLOSEPATH]
        #self.assertEqual(unclosed_path.operators ,[)
        #   _MOVETO, _LINETO, _LINETO, _MOVETO, _LINETO]

    def test_empty_path(self):
        converter = svglib.Svg2RlgShapeConverter(None)
        node = svglib.NodeTracker(parseXML('<path id="W"/>'))
        group = converter.convertPath(node)
        self.assertEqual(group, None)

    def test_clipped_path(self):
        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg version="1.1" xmlns="http://www.w3.org/2000/svg"
                xmlns:xlink="http://www.w3.org/1999/xlink" width="660" height="480">
                <defs>
                    <path id="my-clipping-path" d="M99,176 L 110 170 112 172Z"/>
                    <rect id="my-clipping-rect" x="88.155" y="163" width="419.69" height="20.68"/>
                </defs>
                <clipPath id="my-clip-path">
                    <use xlink:href="#my-clipping-path"/>
                </clipPath>
                <clipPath id="my-clip-rect">
                    <use xlink:href="#my-clipping-rect"/>
                </clipPath>
                <path clip-path="url(#my-clip-path)" d="M99,176 L 110 170 112 172Z"/>
                <path clip-path="url(#my-clip-rect)" d="M99,176 L 110 170 112 172Z"/>
            </svg>
        ''')))
        self.assertIsInstance(drawing.contents[0].contents[0].contents[0], svglib.ClippingPath)
        # Clipping rect was converted to a path
        self.assertIsInstance(drawing.contents[0].contents[1].contents[0], svglib.ClippingPath)


def force_cmyk(rgb):
    c, m, y, k = colors.rgb2cmyk(rgb.red,rgb.green,rgb.blue)
    return colors.CMYKColor(c,m,y,k,alpha=rgb.alpha)

class TestColorAttrConverter(SVGBasicTest):
    "Testing color attribute conversion."

    def test_0(self):
        "Test color attribute conversion."

        mapping = (
            ("red", colors.red),
            ("#ff0000", colors.red),
            ("#f00", colors.red),
            ("rgb(100%,0%,0%)", colors.red),
            ("rgb(255, 0, 0)", colors.red),
        )
        ac = svglib.Svg2RlgAttributeConverter()
        failed = _testit(ac.convertColor, mapping)
        self.assertEqual(len(failed) ,0)

    def test_1(self):
        "Test color attribute conversion to CMYK"

        mapping = (
            ("red", force_cmyk(colors.red)),
            ("#ff0000", force_cmyk(colors.red)),
            ("#f00", force_cmyk(colors.red)),
            ("rgb(100%,0%,0%)", force_cmyk(colors.red)),
            ("rgb(255, 0, 0)", force_cmyk(colors.red)),
            ("rgb(0,255, 0)", force_cmyk(colors.Color(0,1,0))),
            ("rgb(0, 0, 255)", force_cmyk(colors.Color(0,0,1))),
        )
        ac = svglib.Svg2RlgAttributeConverter(color_converter=force_cmyk)
        failed = _testit(ac.convertColor, mapping)
        self.assertEqual(len(failed) ,0)


class TestLengthAttrConverter(SVGBasicTest):
    "Testing length attribute conversion."

    def test_0(self):
        "Test length attribute conversion."

        mapping = (
            ("0", 0),
            ("316", 316),
            ("-316", -316),
            ("-3.16", -3.16),
            ("-1e-2", -0.01),
            ("1e-5", 1e-5),
            ("1e1cm", 10*cm),
            ("1e1in", 10*inch),
            ("1e1%", 10),
            ("-8e-2cm", (-8e-2)*cm),
            ("20px", 20),
            ("20pt", 20 * 1.25),
            ("1.5em", 12 * 1.5),
        )
        ac = svglib.Svg2RlgAttributeConverter()
        failed = _testit(ac.convertLength, mapping)
        self.assertEqual(len(failed) ,0)


    def test_1(self):
        "Test length attribute conversion."

        ac = svglib.Svg2RlgAttributeConverter()
        attr = "1e1%"
        expected = 1
        obj = ac.convertLength(attr, 10)
        self.assertEqual(obj ,expected)


class TestLengthListAttrConverter(SVGBasicTest):
    "Testing length attribute conversion."

    def test_0(self):
        "Test length list attribute conversion."

        mapping = (
            (" 5cm 5in", [5*cm, 5*inch]),
            (" 5, 5", [5, 5]),
        )
        ac = svglib.Svg2RlgAttributeConverter()
        failed = _testit(ac.convertLengthList, mapping)
        self.assertEqual(len(failed) ,0)


class TestTransformAttrConverter(SVGBasicTest):
    "Testing transform attribute conversion."

    def test_0(self):
        "Test transform attribute conversion."

        mapping = (
            ("",
                []),
            ("scale(2) translate(10,-20.5)",
                [("scale", 2.0), ("translate", (10.0, -20.5))]),
            ("scale(0.9), translate(27,40)",
                [("scale", 0.9), ("translate", (27.0, 40.0))]),
            # Invalid expression returns empty list
            ("scale(0.9), translate",
                []),
        )
        ac = svglib.Svg2RlgAttributeConverter()
        failed = _testit(ac.convertTransform, mapping)
        self.assertEqual(len(failed) ,0)


class TestAttrConverter(SVGBasicTest):
    "Testing multi-attribute conversion."

    def test_0(self):
        "Test multi-attribute conversion."

        mapping = (
            ("fill: black; stroke: yellow",
                {"fill":"black", "stroke":"yellow"}),
        )
        ac = svglib.Svg2RlgAttributeConverter()
        failed = _testit(ac.parseMultiAttributes, mapping)
        self.assertEqual(len(failed) ,0)

    def test_findAttr(self):
        """
        Whitespace in attribute values shouldn't disturb parsing.
        """
        ac = svglib.Svg2RlgAttributeConverter()
        node = parseXML('<rect fill=" #00A1DE\n"/>')
        self.assertEqual(ac.findAttr(node, 'fill') ,"#00A1DE")

    def test_findAttr_parents(self):
        ac = svglib.Svg2RlgAttributeConverter()
        rect_node = parseXML(
            '<g style="fill:#008000;stroke:#008000;"><rect style="fill:#ff0;"/></g>'
        ).getchildren()[0]
        self.assertEqual(ac.findAttr(rect_node, 'fill'), "#ff0")
        self.assertEqual(ac.findAttr(rect_node, 'stroke'), "#008000")

    def test_no_fill_on_shape(self):
        """
        Any shape with no fill property should set black color in rlg syntax.
        """
        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg xmlns="http://www.w3.org/2000/svg"
                 width="1200" height="800"
                 viewBox="0 0 36 24">
                <rect y="10" width="36" height="4"/>
            </svg>
        ''')))
        self.assertEqual(drawing.contents[0].contents[0].fillColor ,colors.black)

    def test_fillopacity(self):
        """
        The fill-opacity property set the alpha of the color.
        """
        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg version="1.1"
                 xmlns="http://www.w3.org/2000/svg"
                 width="660" height="480">
                <polygon id="triangle" points="0,-29.14 -25.23, 14.57 25.23, 14.57"
                         stroke="#0038b8" stroke-width="5.5" fill-opacity="0"/>
            </svg>
        ''')))
        self.assertEqual(drawing.contents[0].contents[0].fillColor ,colors.Color(0, 0, 0, 0))

    def test_fillrule(self):
        converter = svglib.Svg2RlgShapeConverter(None)
        node = parseXML('<polygon fill-rule="evenodd"/>')
        poly = Polygon()
        converter.applyStyleOnShape(poly, node)
        self.assertEqual(poly.fillMode ,FILL_EVEN_ODD)

    def test_stroke(self):
        converter = svglib.Svg2RlgShapeConverter(None)
        node = parseXML('<path d="m0,6.5h27m0,5H0" stroke="#FFF" stroke-opacity="0.5"/>')
        path = Path()
        converter.applyStyleOnShape(path, node)
        self.assertEqual(path.strokeColor ,colors.white)
        self.assertEqual(path.strokeOpacity, 0.5)
        self.assertEqual(path.strokeWidth ,1)


class TestApplyTransformOnGroup(SVGBasicTest):
    def test_translate_only_x(self):
        """
        When the second translate value is missing, 0 is assumed.
        """
        group = Group()
        converter = svglib.Svg2RlgShapeConverter(None)
        transform = "translate(10)"
        converter.applyTransformOnGroup(transform, group)
        self.assertEqual(group.transform ,(1, 0, 0, 1, 10, 0))


class TestTextNode(SVGBasicTest):
    def test_font_family(self):
        def font_config_available():
            try:
                subprocess.call(["fc-match"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            except OSError:
                return False
            return True

        converter = svglib.Svg2RlgAttributeConverter()
        # Check PDF standard names are untouched
        self.assertEqual(converter.convertFontFamily('ZapfDingbats'), 'ZapfDingbats')
        self.assertEqual(converter.convertFontFamily('bilbo ZapfDingbats'), 'ZapfDingbats')
        self.assertEqual(converter.convertFontFamily(' bilbo      ZapfDingbats  '), 'ZapfDingbats')
        self.assertEqual(converter.convertFontFamily(' bilbo,    ZapfDingbats  '), 'ZapfDingbats')
        if font_config_available():
            # Fontconfig will always provide at least a default font and register
            # that font under the provided font name.
            self.assertEqual(converter.convertFontFamily('SomeFont'), 'SomeFont')
        else:
            # Unknown fonts are converted to Helvetica by default.
            self.assertEqual(converter.convertFontFamily('SomeFont'), 'Helvetica')

    def test_space_preservation(self):
        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg width="777" height="267">
              <text style="fill:#000000; stroke:none; font-size:28;">
                <tspan>TITLE    1</tspan>
                <tspan x="-10.761" y="33.487">Subtitle</tspan>
              </text>
            </svg>
        ''')))
        main_group = drawing.contents[0]
        # By default, only two tspans produce String objects, the rest
        # (spaces/newlines) is ignored.
        self.assertEqual(len(main_group.contents[0].contents) ,2)
        self.assertEqual(main_group.contents[0].contents[0].text ,"TITLE 1")

        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg width="777" height="267" xml:space="preserve">
              <text style="fill:#000000; stroke:none; font-size:28;">%s<tspan>TITLE    1</tspan><tspan x="-10.761" y="33.487">Subtitle</tspan></text>
            </svg>
        ''' % (5*' '))))
        main_group = drawing.contents[0]
        self.assertEqual(main_group.contents[0].contents[0].text ,5*' ')
        self.assertEqual(main_group.contents[0].contents[1].text ,"TITLE    1")

        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg width="777" height="267">
              <text style="fill:#000000; stroke:none; font-size:28;">
                <tspan>TITLE    1</tspan>
                <tspan x="-10.761" y="33.487">Subtitle</tspan>
              </text>
              <text xml:space="preserve">  with   spaces </text>
            </svg>
        ''')))
        main_group = drawing.contents[0]
        # xml:space can be overriden per text node
        self.assertEqual(main_group.contents[0].contents[0].text ,"TITLE 1")
        self.assertEqual(main_group.contents[1].contents[0].text ,'  with   spaces ')

    def test_tspan_position(self):
        """
        The x/y positions of a tspan are either relative to the current text
        position, or can be absoluted by specifying the x/y attributes.
        """
        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg width="777" height="267">
              <text x="10" y="20" style="fill:#000000; stroke:none; font-size:28;">
                <tspan>TITLE 1</tspan>
                <!-- x position relative to current text position
                     y position offset in em -->
                <tspan dy="1.3em">(after title)</tspan>
                <!-- absolute position -->
                <tspan x="16.75" y="33.487">Subtitle</tspan>
                <!-- absolute position + shifting -->
                <tspan x="10" y="20" dx="3em" dy="1.5em">Complete</tspan>
              </text>
            </svg>
        ''')))
        main_group = drawing.contents[0]
        self.assertEqual(main_group.contents[0].contents[0].x, 10)
        self.assertEqual(main_group.contents[0].contents[0].y, -20)
        self.assertTrue(main_group.contents[0].contents[1].x > 10)
        self.assertEqual(main_group.contents[0].contents[1].y, -20 - (1.3 * 28))
        self.assertEqual(main_group.contents[0].contents[2].x, 16.75)
        self.assertEqual(main_group.contents[0].contents[2].y, -33.487)
        self.assertEqual(main_group.contents[0].contents[3].x, 10 + (3 * 28))
        self.assertEqual(main_group.contents[0].contents[3].y, -20 - (1.5 * 28))


class TestPolylineNode(SVGBasicTest):
    def test_filling(self):
        converter = svglib.Svg2RlgShapeConverter(None)
        node = svglib.NodeTracker(parseXML(
            '<polyline fill="none" stroke="#000000" '
            'points="10,50,35,150,60,50,85,150,110,50,135,150" />'
        ))
        polyline = converter.convertPolyline(node)
        self.assertIsInstance(polyline, PolyLine)

        # svglib simulates polyline filling by a fake polygon.
        node = svglib.NodeTracker(parseXML(
            '<polyline fill="#fff" stroke="#000000" '
            'points="10,50,35,150,60,50,85,150,110,50,135,150" />'
        ))
        group = converter.convertPolyline(node)
        self.assertIsInstance(group.contents[0], Polygon)
        self.assertEqual(group.contents[0].fillColor ,colors.white)
        self.assertIsInstance(group.contents[1], PolyLine)


class TestUseNode(SVGBasicTest):
    def test_use(self):
        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg version="1.1"
                 xmlns="http://www.w3.org/2000/svg"
                 xmlns:xlink="http://www.w3.org/1999/xlink"
                 width="10cm" height="3cm" viewBox="0 0 100 30">
              <defs>
                  <rect id="MyRect" width="60" height="10"/>
              </defs>
              <rect x=".1" y=".1" width="99.8" height="29.8"
                    fill="none" stroke="blue" stroke-width=".2" />
              <use x="20" y="10" xlink:href="#MyRect" />
              <use x="30" y="20" xlink:href="#MyRect" fill="#f00" />
            </svg>
        ''')))
        main_group = drawing.contents[0]
        # First Rect
        self.assertIsInstance(main_group.contents[0], Rect)
        # Second Rect defined by the use node (inside a Group)
        self.assertIsInstance(main_group.contents[1].contents[0], Rect)
        self.assertEqual(main_group.contents[1].contents[0].fillColor ,colors.black) #default
        # Attributes on the use node are applied to the referenced content
        self.assertIsInstance(main_group.contents[2].contents[0], Rect)
        self.assertEqual(main_group.contents[2].contents[0].fillColor ,colors.red)

    def test_transform_inherited_by_use(self):
        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg version="1.1"
                 xmlns:xlink="http://www.w3.org/1999/xlink"
                 width="900" height="600">
                <g id="c">
                    <path id="t" d="M 0,-100 V 0 H 50"
                          transform="rotate(18 0,-100)"/>
                    <use xlink:href="#t" transform="scale(-1,1)"/>
                </g>
            </svg>
        ''')))
        cgroup_node = drawing.contents[0].contents[0]
        self.assertEqual(
            cgroup_node.contents[0].transform,cgroup_node.contents[1].contents[0].transform
            )

    def test_use_forward_reference(self):
        """
        Sometimes, a node definition pointed to by xlink:href can appear after
        it has been referenced. But the order should remain.
        """
        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg version="1.1"
                 xmlns:xlink="http://www.w3.org/1999/xlink"
                 width="900" height="600">
                <use xlink:href="#back" x="-100"/>
                <rect id="back" x="42" y="42" width="416" height="216" fill="#007a5e"/>
            </svg>
        ''')))
        self.assertEqual(len(drawing.contents[0].contents) ,2)
        self.assertIsInstance(drawing.contents[0].contents[0], Group)
        self.assertIsInstance(drawing.contents[0].contents[1], Rect)

    def test_use_node_properties(self):
        """
        Properties on the use node apply to the referenced item.
        """
        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg version="1.1"
                 xmlns:xlink="http://www.w3.org/1999/xlink"
                 width="900" height="600">
                <path id="a" fill="#FF0000" d="M-15 37.57h60L-15 0v80h60l-60-60z"/>
                <use stroke="#003893" stroke-width="5" xlink:href="#a"/>
                <use stroke="#003893" stroke-width="2" xlink:href="#a"/>
            </svg>
        ''')))
        use_path1 = drawing.contents[0].contents[1].contents[0].contents[0]
        use_path2 = drawing.contents[0].contents[2].contents[0].contents[0]
        # Attribute from <path> node
        self.assertEqual(use_path1.fillColor ,colors.Color(1, 0, 0, 1))
        # Attribute from <use> node
        self.assertEqual(use_path1.strokeWidth ,5)
        self.assertEqual(use_path2.strokeWidth ,2)

    if 0:
        def test_use_node_with_unclosed_path(self):
            """
            When a <use> node references an unclosed path (which is a group with two
            different paths for filling and stroking), the use properties shouldn't
            affect the no-stroke property of the fake stroke-only path.
            """
            drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
                <?xml version="1.0"?>
                <svg xmlns="http://www.w3.org/2000/svg"
                     xmlns:xlink="http://www.w3.org/1999/xlink"
                     width="900" height="600" viewBox="0 0 9 6">
                    <defs>
                        <path d="M0,0 4.5,3 0,6" id="X"/>
                    </defs>
                    <use xlink:href="#X" fill="#f00" stroke="#ffb612" stroke-width="2"/>
                </svg>
            ''')))
            use_group = drawing.contents[0].contents[0].contents[0]
            self.assertEqual(use_group.contents[0].getProperties()['strokeWidth'] ,0)
            self.assertEqual(use_group.contents[0].getProperties()['strokeColor'], None)

class TestSymbolNode(SVGBasicTest):
    def test_symbol_node(self):
        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg xmlns="http://www.w3.org/2000/svg"
                 xmlns:xlink="http://www.w3.org/1999/xlink"
                 width="900" height="600" viewBox="0 0 100 100">
                <defs>
                    <symbol id="X">
                        <path d="M 0,10 V 100 H 50 Z"/>
                    </symbol>
                </defs>
                <use xlink:href="#X" fill="#f00"/>
            </svg>
        ''')))
        self.assertIsInstance(drawing.contents[0].contents[0].contents[0].contents[0].contents[0],svglib.Path)

class TestViewBox(SVGBasicTest):
    def test_nonzero_origin(self):
        drawing = svglib.svg2rlg(io.StringIO(textwrap.dedent(u'''\
            <?xml version="1.0"?>
            <svg xmlns="http://www.w3.org/2000/svg"
                 width="1200" height="800"
                 viewBox="-60 -40 120 80">
                <g fill="#E70013">
                    <rect x="-60" y="-40" width="120" height="80"/>
                </g>
            </svg>
        ''')))
        # Main group coordinates are translated to match the viewBox origin
        self.assertEqual(drawing.contents[0].transform ,(1, 0, 0, -1, 60.0, 40.0))

def makeSuite():
    return makeSuiteForClasses( TestPaths, TestColorAttrConverter, TestLengthAttrConverter, TestLengthListAttrConverter,
                                    TestTransformAttrConverter, TestAttrConverter, TestApplyTransformOnGroup,
                                    TestTextNode, TestPolylineNode, TestUseNode, TestSymbolNode, TestViewBox,
                )

if __name__ == "__main__":
    import sys
    w = os.path.dirname(sys.argv[0])
    if w: os.chdir(w)
    unittest.TextTestRunner().run(makeSuite())



__code__='Gas1]gMYJ*&Daab91rqc.=D;$gk:f/H5AmR!GOfadSLMuaFGhn]5?@b[_\'.c;lb==?hk7^afkUG:X$=?)m42SBY6-BAnO!AhoXO>)Q`,+A-ftg392r*P`[c=a?XepVZYr*r_\'CKNTnYp"]f.[<b<"]P$MK2G*Pe,Bi\\fs^/".]"0tD1b^S+__^Y(!Vqg3aJPO>\'>).J\\4r)]Y]7*5>&t<_OCG)A&J,C_kkU:KP*6L%rBM]<KT2A*k\'TrfTK!XS)_]30rlBo&</^P.l7o"1q:<8:;["0hG(!US9O.rup/[9@M*G>OITi\'`*>3QdrSocpKY.*7Fem?%roW5F;Q)Bkf8oeSZ&eROS1,;5fE\'\'glVK_-^UMWHWLaYX`+-m1X?O?2KP.bZbXjj09CW!8X:BdGnosWkAfu\\62,me`<0os[:Xh^,=*BQhDg4C5Sf8[EqGMR)`Fh9BoKkOOmpRp]\'k8l>aWHmGWQpAGkdZ>KC9[aIB/PfiLRbj@@fmM[8")3\'n)#-4VS=,36IUAN5p/p+Q;TY,+mHEcT(AVS+D:sBn[E[3=*-u&>;aDCkYqUGf#mAWMEH:t-#M+M%#PD2T?D9n(ruGg6,6&O#kI41r\\WD%G(!!305d)<op=KD$9RdU*f&K;Hd.O6u_&pp*^hY8[iU;)UgWn=.r5(So<u4o\'=-ju+1RGZ<Ol@SU_Xr<(`P"oJXOeFMmR_J5VH&,4qti[d1V/inAop22$b[q:fo%AuBQMo96I+$$d;HUhcEt.>\'7i2&56PtnpM<PcqMTM9T[e,mJ9.gT\\:M0;7->o47LS0<Cc4K._5sE0fmrg/r=tc>OR\'f<<,o:&,K\\aFkhf9KO.u2@;-lZPiWF17T_rin>[TP>q,H?YZbTWr]dR-:")?5Bp4"%?XQ\\e]]8:*X]PrR^.]VC#3KmIg9!7K6+].:E#4cA[5k2#$k^Gu3"Y)\'!qO:*_o$_Yc4i;C;&AKN5W`Z&$ec,0aC/o"^563;R>bD~>'



#Autogenerated by ReportLab guiedit do not edit
from reportlab.graphics.charts.piecharts import Pie
from reportlab.graphics.shapes import Drawing, _DrawingEditorMixin
from reportlab.lib.colors import black

class Drawing_000(_DrawingEditorMixin,Drawing):
	def __init__(self,width=400,height=200,*args,**kw):
		Drawing.__init__(self,width,height,*args,**kw)
		self._add(self,Pie(),name=None,validate=None,desc=None)
		self.contents[0].slices[0].strokeColor        = black
		self.contents[0].slices[0].strokeWidth        = 1

if __name__=="__main__": #NORUNTESTS
	Drawing_000().save(formats=['pdf'],outDir='.',fnRoot=None)



#!/usr/local/bin/python
"""Script to check for bad UTF-8 in a MySQL database.
NOT dependent on any assumptions about Django.

Tested on:
 - MySQL 5.0
"""
from __future__ import print_function
import sys

def getConnection(db='',host='',user='',passwd=''):
    import MySQLdb
    conn = MySQLdb.connect(
        host=host,
        user=user,
        passwd=passwd,
        db=db,
        charset='utf8',
        use_unicode=False,
        )
    return conn

def parseCommandLine():
    """Examines options and does preliminary checking"""
    from optparse import OptionParser

    parser = OptionParser(usage="%prog [options] database....",version='1.0')
    parser.add_option("-v", "--verbose",
                      action="count", dest="verbose", default=0,
                      help="Print lots of information about what is going on")
    parser.add_option("--host",
                      dest="host",
                      default='localhost',
                      help="hostname (localhost)")
    parser.add_option("-u","--user",
                      dest="user",
                      default='root',
                      help="user name (root)")
    parser.add_option("-p","--password",
                      dest="passwd",
                      default='',
                      help="user name (root)")
    options,args = parser.parse_args()
    if not len(args):
        parser.error("%prog needs at least one database argument")
    return options,args

def checkdb(options,db):
    conn = getConnection(db=db,host=options.host,user=options.user,passwd=options.passwd)
    cur = conn.cursor()
    try:
        cur.execute("""\
SELECT t.TABLE_NAME, c.COLUMN_NAME, c.COLUMN_TYPE FROM information_schema.TABLES t, information_schema.COLUMNS c
  where t.TABLE_SCHEMA = c.TABLE_SCHEMA
  and t.TABLE_NAME = c.TABLE_NAME
  and (c.COLUMN_TYPE like '%%char%%' or c.COLUMN_TYPE like '%%text%%')
  and t.TABLE_SCHEMA = %s
  order by t.TABLE_NAME, c.COLUMN_NAME, c.COLUMN_TYPE
""",
                (db,),
                )
        columnRows = cur.fetchall()
    except:
        try:
            import re
            pat=re.compile('^.*(?:text|char).*$',re.I)
            cur.execute("""show table status from `%s`""" % db)
            T = [t[0] for t in cur.fetchall()]
            columnRows = []
            for tableName in T:
                cur.execute("""show columns from `%s`.`%s`""" % (db,tableName))
                data = cur.fetchall()
                columnRows += [(tableName,x[0],x[1]) for x in data if pat.match(x[1])] 
        except:
            print("""Could not get schema as per modern MySQL and older version failed as well""", file=sys.stderr)
            raise
    verbose = options.verbose
    if verbose:
        print(db)
    if verbose>2:
        print(columnRows)
    allRows = [None,None]
    for tableName, colName, colType in columnRows:
        if verbose>1:
            print(tableName, colName, colType)
        cur.execute("SELECT `%s` FROM %s.`%s`" % (colName, db, tableName))
        dataRows = cur.fetchall()
        for ii, dataRow in enumerate(dataRows):
            data = dataRow[0]
            if data is None: continue
            try:
                data.decode('utf-8')
            except UnicodeError:
                # we can't assume when doing the dataRows query that the first
                # primary key is called 'id' since that's not true in general
                # even of Django's primary keys (and MySQL < 5.1 doesn't tell us
                # which is the primary key), so select * now so we can report
                # which row we're looking at if we find a problem
                if allRows[0]!=tableName:
                    cur.execute("SELECT * FROM `%s`.`%s`" % (db, tableName))
                    allRows[0]=tableName
                    allRows[1]=cur.fetchall()
                if verbose>1:
                    print()
                print('%s.`%s`.`%s`:\n%s\n%s\n\n' % (
                    db, tableName, colName, data, allRows[1][ii][0]))
            if verbose>1 and not ii%10:
                print('.', end=' ')
        if verbose>1: print()

def main():
    options,args = parseCommandLine()
    for db in args:
        checkdb(options,db)


if __name__ == '__main__':
    main()



#Autogenerated by ReportLab guiedit do not edit
from reportlab.graphics.charts.textlabels import Label
from reportlab.graphics.charts.barcharts import HorizontalBarChart
from reportlab.graphics.shapes import _DrawingEditorMixin, String
from rlextra.graphics.guiedit.datacharts import DataAwareDrawing, ODBCDataSource, DataAssociation

class SampleDataChart(_DrawingEditorMixin,DataAwareDrawing):
	def __init__(self,width=400,height=200,*args,**kw):
		DataAwareDrawing.__init__(self,width,height,*args,**kw)
		self._add(self,HorizontalBarChart(),name='chart',validate=None,desc=None)
		self.chart.x                = 75
		self.chart.y                = 30
		self._add(self,String(125, 124, 'I am a string'),name='title',validate=None,desc=None)
		self.title.x              = 125
		self.title.y              = 125
		self.dataSource      = ODBCDataSource()
		self.dataSource.sql            = 'SELECT chartId, rowId, name, value1, value2, value3 FROM generic_bar'
		self.dataSource.associations.size = 4
		self.dataSource.associations.element00 = DataAssociation(column=0, target='chartId', assocType='scalar')
		self.dataSource.associations.element01 = DataAssociation(column=2, target='title.text', assocType='scalar')
		self.dataSource.associations.element02 = DataAssociation(column=[3,4,5], target='chart.data', assocType='tmatrix')
		self.dataSource.associations.element03 = DataAssociation(column=2, target='chart.categoryAxis.categoryNames', assocType='vector')

if __name__=="__main__": #NORUNTESTS
	SampleDataChart().go()	



__code__='Gas1]D/U(X&>$Qs4,unKR$<mPb*A;/=6^Y/TW5#l,%FFpa,Md7l=b=Ogh`,ukIY5BSN*P:.7jEc5LR<oT5e680D_Y((E[XjMS2)Tn\'4&Ab]ocY)s0=cC&ULVqhkkh#8c;)$:\'-N_"PhL=+D!("MC-$%ak?Q@00V_28D92!al+L&[2Rd#&jn<,J^X9k5\\_:k5j=f(BBd&56eb+p^,lVJ<p7#"8j/.1>in_0*&0gbIL\\$F$9Be1Ajc"(:Z5`Fj<6U+DYq,Jp@5!CqR.5-*c3T?ZN/!j_tYuSK!:8#j)LOWrfh0=Va;7\\Q"-IeaZ[jC\\#eKa&7N%L4M>m[Uukd!3=lu!J^_/^+hnU*7?MYKQ>63TLH7uQ<@r)%LkrN<j%%DGpGWd5>_m]1Z$C#>CX4FI+Ef>pfYWg(:L(5[8t752I"[/\'umgp\'ilZVX[9s[3R5\'=@l+)t<.bh[VkYFre]Ro8`OY05h!\\i)kg$WJ\'-0PFGI!W)kbJH;OeuAc`c3,iU+KOm[l0a5-%^m7`Vrtc?IADXEIsC1f1V_0HK%H8-`TaX-0q((AA_\\/(n/^4E6X05;8-X,Q"cub^u)NN#L>n`R+]1oKn51=M0>nT,g^_?2<VHDFo$Nq]>),LW,a]b2k;\'/4#ULW#J&=@o:q/G_^)Z(A]nd+.`)OUo\\dJL9LqGb"V($,]gShkW[mYhrHl(\\q4Hr]Q=#-F^ceYPc:fUBXX$d.A(S+SIVg_-)<qI"<8i.P\\!P%eVnSZ@htiud*MTdeM$eG(9KElt2K[:%]m]\'%N<a[qhg;Tu*H+-fgF"DMZ\'Hm5,>+p79/$;p*:q8Y%4<ltouGDFfof`hK=hrcaC$rc\\#U"ep7S%eX+"+L_p+.e#JJ1Ro".Y<TY9HBO[K]U3H%qgaY)J7(O:?QK8Bs.Jc-cdU9[5V\'[.Ao?4"t%1]<d(oLb>Y`#A[l);.WCX>#8JLaHr&.MK^fVYP[N(hgM"/83!=[3H#5IIAX%Fg%%b\'dg<1\'O!i2pf6O6&uj>.(-Kp(,(_;,/]KAWTpF`?&A*YK;_571kl9.11k%F8f8goCID[@g$hZrn.-d^X0,_s-DYsm`CLjg~>'



#Copyright ReportLab Europe Ltd. 2000-2017
#see license.txt for license details
from __future__ import print_function
__version__='3.3.0'
import os, sys, string, traceback, re, glob, struct, time
from shutil import copy2
from reportlab.lib.utils import rl_exec
path_join = os.path.join
path_basename = os.path.basename
path_dirname = os.path.dirname
isFile = os.path.isfile
isDir = os.path.isdir
isLink = os.path.islink
normPath = os.path.normpath
os_sep = os.sep

try:
	set
except NameError:
	from sets import Set as set	#backward compatibility

if hasattr(sys, 'version_info'):
	_pyvers = sys.version_info[0]*10 + sys.version_info[1]
else:
	toks = string.split(sys.version, '.', 2)
	_pyvers = int(toks[0])*10 + int(toks[1])

def errorMsg(msg,exit=1):
	print(msg, file=sys.stderr)
	if exit is not None: sys.exit(exit)

def getFlag(name,nextVal=0,default=None):
	r = name in sys.argv
	if nextVal and not r:
		name += '='
		for x in sys.argv[1:]:
			if x.find(name)==0:
				name = x
				r = '='.join(name.split('=')[1:])
				nextVal = 0
				break
	if r:
		x = sys.argv.index(name)
		del sys.argv[x]
		if nextVal:
			try:
				r = sys.argv[x]
				del sys.argv[x]
			except:
				r = (nextVal>>1)&1
	else:
		r = default
	return r
_verbose = os.environ.get('RL_verbose',0)

def invalidFlags(allowed,trailersOK=0):
	I = []	#invalid flags
	for a in sys.argv[1:]:
		if a[0]=='-':
			if a[:2]=='--' and a.find('=')>0: a = a.split('=')[0]
		elif trailersOK: continue
		if a not in allowed: I.append(a)
	return I

class STDLOG:
	def __init__(self,fn='log.txt'):
		self.log = [sys.__stdout__,open(path_join(path_dirname(sys.argv[0]),fn),'w')]
		sys.stderr = sys.stdout = self

	def write(self,msg):
		for log in self.log:
			log.write(msg)

	def close(self):
		sys.stderr = sys.__stderr__
		sys.stdout = sys.__stdout__
		self.log[1].close()

	def __del__(self):
		self.close()

def installerDir(P=None):
	import rlextra
	D = path_join(path_dirname(rlextra.__file__),'utils','installer')
	if P:
		if type(P) not in (type(()),type([])):
			D = path_join(D,P)
		else:
			for p in P:
				D = path_join(D,p)
	return normPath(D)

def do_exec(cmd, fout=None, sys_exit=1, verbose=None, fail=1, report=None):
	if verbose is None: verbose = _verbose
	i=os.popen(cmd,'r')
	o = i.read()
	if fout: print(o, file=fout)
	i = i.close()
	if fail and i is not None:
		errorMsg('Error: %s\n%s\n' % (cmd,o),sys_exit)
		raise ValueError('external command failure')
	else:
		if verbose>8: print('do_exec(%s)' % repr(cmd), file=sys.stderr)
		if report: report(cmd)

def get_path():
	p = os.environ.get('PATH','')
	if not p:
		for k,v in os.environ.items():
			if k.upper()=='PATH':
				p = v
				break
	return p.split(os.pathsep)

_find_exe_cache = {}
def find_exe(name, doRaise=0, known_places=[]):
	global _find_exe_cache
	if sys.platform=='win32':
		name = '%s.exe' % name

	exe = _find_exe_cache.get(name,None)
	if exe: return exe

	for p in known_places+get_path():
		f = path_join(p,name)
		if isFile(f):
			if ' ' in f: f = '"%s"' % f
			_find_exe_cache[name] = f
			return f

	msg = "Can't find %s anywhere on the path" % name
	if doRaise: raise RuntimeError(msg)
	errorMsg(msg)

def find_exe_nq(name,*args,**kwds):
	exe = find_exe(name,*args,**kwds)
	if ' ' in exe and exe.startswith('"'):
		exe = exe[1:-1]
	return exe

def patchRead(patchFn,filterFunc=None):
	patchFile = open(patchFn,'r')
	patchText = patchFile.read()
	patchFile.close()
	P = []
	C = []
	for i,l in enumerate(patchText.split('\n')):
		if l.startswith('#'): continue
		if l.startswith('---'):
			fn = l.split()
			if len(fn):
				fn = fn[1]
			else:
				raise ValueError('malformed patch line%03d: "%s"' %(i+1,l))
			if C:
				P.append((C[0],'\n'.join(C[1:])))
				C = []
			C.append(fn)
		elif not C:
			raise ValueError('missing patch filename at line%03d: "%s"' %(i+1,l))
		else:
			C.append(l)
	if C:
		P.append((C[0],'\n'.join(C[1:])))
	if filterFunc:
		P = list(filter(filterFunc,P))
	return P

def patchRestore(patchFn,P=None,filterFunc=None):
	if P is None:
		P = patchRead(patchFn,filterFunc=filterFunc)
	out = []
	for x,(fn,p) in enumerate(P):
		ofn = fn+'.org'
		if isFile(ofn):
			kill(fn)
			os.rename(ofn,fn)
			out.append('Restored '+fn)
	return '\n'.join(out)

def patch(patchFn,testOnly=0,keepOld=1,filterFunc=None):
	exe = find_exe('patch',doRaise=1)
	P = patchRead(patchFn,filterFunc=filterFunc)
	try:
		out = []
		for x,(fn,p) in enumerate(P):
			if not isFile(fn):
				out.append('!!! not patching missing file '+fn)
				continue
			ofn = fn+'.out'
			rfn = fn+'.rej'
			kill(ofn)
			kill(rfn)
			cmd = '%s --force -o "%s" -r "%s" "%s"' % (exe,ofn,rfn,fn)
			i,oe = os.popen4(cmd,'t')
			i.write('\n'+p+'\n')
			i.close()
			outerr = oe.read()
			oe.close()
			kill(fn+'.orig')		#in case patch created an unwanted one
			kill(fn+'.out.orig')	#in case patch created an unwanted one
			if isFile(ofn) and not isFile(rfn):
				#success
				if not testOnly:
					if keepOld:
						gfn = fn+'.org'
						kill(gfn)
						os.rename(fn,gfn)
					else:
						kill(fn)
					os.rename(ofn,fn)
				out.append('#### patched '+fn)
				continue
			out.append('!!!!!!!!!!!!!!!!!!!!\n	  failed to patch '+fn)
			out.append('	'+cmd)
			if not isFile(ofn): out.append('file "%s" not created' % ofn)
			if isFile(rfn): out.append('rej file created')
			out.append(outerr.rstrip())
			out.append('!!!!!!!!!!!!!!!!!!!!')
			if not testOnly:
				err = 'Cannot patch "%s" using\n--------------\n%s\n--------------'%(fn,p)
				raise ValueError(err)
			kill(ofn)
			kill(rfn)
	except:
		if keepOld and not testOnly:
			out.append(patchRestore(patchFn,P=P[:x+1],filterFunc=filterFunc))
		t,e,tb = sys.exc_info()
		raise t(t('\n'.join(list(e.args)+out))).with_traceback(tb)
	return '\n'.join(out)

def http_ref_fix(src,fixes,dst=None):
	'''
	src is a folder to be fixed, fixes is a dictionary
	'src relative filename': [http refs]

	for each file
		1) check it exists
		2) replace the http://ref with /ref
		3) ensure the file referenced is created as dst/ref
		4) ensure no http://refs remain in the file
	'''
	import io, traceback, urllib.request, urllib.parse, urllib.error
	if not dst: dst = src
	out = io.StringIO()
	err = io.StringIO()
	for fn,L in fixes.items():
		try:
			fn = path_join(src,fn)
			f = open(fn,'r')
			text = f.read()
			f.close()
			for ref in L:
				if ref.startswith('http://'):
					rref = ref[6:]
				elif ref.startswith('https://'):
					rref = ref[7:]
				text = text.replace(ref,rref)
				dfn = path_join(dst,rref[1:].replace('/',os.sep))
				if not isFile(dfn):
					safe_makedirs(path_dirname(dfn))
					f = open(dfn,'wb')
					f.write(urllib.request.urlopen(ref).read())
					f.close()
					print('Copied',ref,'to',dfn, file=out)
			f = open(fn,'w')
			f.write(text)
			f.close()
		except:
			print('http_fix: Error processing file', fn, file=err)
			traceback.print_exc(file=err)
			continue
		print('http_fix: processed',fn, file=out)
	return out.getvalue(), err.getvalue()

def ext_remove(d,ext):
	'remove .ext n d and subtree'
	if isDir(d):
		for p in os.listdir(d):
			fn = path_join(d,p)
			if isFile(fn) and fn[-len(ext):]==ext:
				os.remove(fn)
			else:
				ext_remove(fn,ext)

def rmdir(d):
	'destroy directory d'
	if isDir(d):
		try:
			os.chmod(d,0o777)
		except:
			pass
		for p in os.listdir(d):
			fn = path_join(d,p)
			if isDir(fn):
				rmdir(fn)
			else:
				remove(fn)
		os.rmdir(d)

def remove(f):
	'destroy an existing file'
	try:
		if not isLink(f):
			try:
				os.chmod(f,0o666)
			except:
				pass
		os.remove(f)
	except OSError:
		pass

def _kill(f):
	'remove a single directory or file unconditionally'
	if isFile(f): remove(f)
	elif isDir(f): rmdir(f)

def kill(f):
	'remove directory or file unconditionally'
	list(map(_kill,glob.glob(f)))

def filesMatch(patterns):
	L = []
	Lappend = L.append
	for p in patterns:
		G = glob.glob(p)
		while G:
			g = G[-1]
			del G[-1]
			if isDir(g):
				G.extend(listAllFiles(g,rel=0))
			else:
				Lappend(g.replace(os.sep,'/'))
	return L

def copyTree(src, dst, symlinks=0,excludeSVN=1,exclude=[]):
	"""Recursively copy a directory tree using copy2."""
	if not isDir(dst): os.mkdir(dst)
	for f in os.listdir(src):
		if f in exclude: continue
		s = path_join(src, f)
		d = path_join(dst, f)
		try:
			if symlinks and os.path.islink(s):
				linkto = os.readlink(s)
				os.symlink(linkto, d)
			elif isDir(s):
				if not (excludeSVN and path_basename(s)=='.svn'): copyTree(s, d, symlinks,excludeSVN)
			else:
				if not (excludeSVN and path_basename(s)[:4]=='.svn'): copy2(s, d)
		except (IOError, os.error) as why:
			print("Can't copy %s to %s: %s" % (repr(s), repr(d), str(why)))
			raise


def safe_umask(mask):
	'''non erroring umask'''
	if mask is None: return None
	try:
		return os.umask(mask)
	except:
		return None

def safe_makedirs(dir,mode=0o777):
	'''non erroring makedirs'''
	try:
		um = safe_umask(0)
		os.makedirs(dir,mode)
	except:
		pass
	safe_umask(um)

def _src2dst(src,dst,excludeSVN=1):
	if isFile(src):
		if _verbose>1: print(' %s --> ' % src, end=' ')
		D = path_dirname(dst)
		if not isDir(D): os.makedirs(D)
		copy2(src,dst)
		if _verbose>1: print(dst)
	elif isDir(src):
		if _verbose>1: print(' %s --> ' % src, end=' ')
		if not isDir(dst): os.makedirs(dst)
		copyTree(src,dst,excludeSVN=excludeSVN)
		if _verbose>1: print(dst)

def copyOver(L,src,dst,xtra=None,excludeSVN=1):
	assert type(L) in (type(()),type([])), 'L should be list or tuple'
	for x in L:
		d = path_join(dst,x)
		_src2dst(path_join(src,x),d,excludeSVN)
		if xtra: _src2dst(path_join(xtra,x),d,excludeSVN)	#overwrite with any specials

_bpath=None
_ppath=None
def getExpandedPath():
	"""Return sys.path as expanded by packages"""
	global _ppath
	if _ppath is None:
		raise NotImplementedError('_ppath = expand(sys.path)')
	return _ppath

def _locate(nm, xtrapath=None, base=None):
	"""Find a file / directory named NM in likely places.
	
	   XTRAPATH is a list of paths to prepend to BASE.
	   If BASE is None, sys.path (as extended by packages) is used."""
	ppath = base
	if base is None:
		ppath = getExpandedPath()
	if xtrapath:
		ppath = xtrapath + ppath
	for pth in ppath:
		fullnm = path_join(pth, nm)
		#print " _locate trying", fullnm
		if os.path.exists(fullnm):
			break
	else:
		return ''
	return fullnm

def getWindowsPath():
	"""Return the path that Windows will search for dlls."""
	global _bpath
	if _bpath is None:
		try:
			import win32api
		except ImportError:
			print("W: Cannot determine your Windows or System directories")
			print("W: Please add them to your PATH if .dlls are not found")
			_bpath = []
		else:
			sysdir = win32api.GetSystemDirectory()
			sysdir2 = path_join(sysdir, '../SYSTEM')
			windir = win32api.GetWindowsDirectory()
			_bpath = [sysdir, sysdir2, windir]
		_bpath.extend(string.split(os.environ.get('PATH', ''), ';'))
	return _bpath

class SpecialDistros:
	def pythonDistro(self,full=0,postcludes=[]):
		from rlextra.utils.superglob import FileSet
		lib = _locatePythonLib()
		fs = FileSet(lib)
		fs.include('./**/*')
		fs.exclude('./site-packages/**/*')
		fs.exclude('./**/*.pyc')
		fs.exclude('./**/*.pyo')
		if not full:
			fs.exclude('./test/**/*')
			fs.exclude('./email/test/**/*')
			fs.exclude('./idlelib/**/*')
			fs.exclude('./distutils/**/*')
			fs.exclude('./lib-old/**/*')
			fs.exclude('./lib-tk/**/*')
		for d in postcludes: getattr(fs,d[0])(d[1])
		return fs

	def pilDistro(self,full=1,postcludes=[],dlls=0):
		from rlextra.utils.superglob import FileSet
		lib = _locatePythonLib()
		fs = FileSet(path_join(lib,'site-packages','PIL'))
		if not dlls:
			fs.include('./**/*.py')
		else:
			if sys.platform=='win32':
				ext = '.pyd'
			else:
				ext = '.so'
			fs.include('./**/*'+ext)
		for d in postcludes: getattr(fs,d[0])(d[1])
		return fs

	def pythonDLLs(self,postcludes=[]):
		from rlextra.utils.superglob import FileSet, SuperFileSet
		fs = FileSet(path_join(sys.exec_prefix,'DLLs'))
		fs.include('./*.pyd')
		#fs.include('./*.dll')
		F = [fs]
		for fn in (_locateVerDLL('Python'), _locateVerDLL('PyWinTypes')):
			fs = FileSet(path_dirname(fn))
			fs.include('./'+path_basename(fn))
			F.append(fs)
		sfs = SuperFileSet(F)
		for d in postcludes: getattr(sfs,d[0])(d[1])
		return sfs

	def reportlabDLLs(self,postcludes=[]):
		from rlextra.utils.superglob import FileSet, SuperFileSet
		F = []
		for modname in '''_renderPM _rl_accel sgmlop pyRXPU pyRXP pyHnj'''.split():
			NS = {}
			rl_exec("import %s as mod" % modname,NS)
			mod = NS['mod']
			del NS
			fn = mod.__file__
			fs = FileSet(path_dirname(fn))
			fs.include('./'+path_basename(fn))
			F.append(fs)
		sfs = SuperFileSet(F)
		for d in postcludes: getattr(sfs,d[0])(d[1])
		return sfs

def _locatePythonLib():
	if sys.platform=='win32':
		return path_join(sys.exec_prefix,'Lib')
	else:
		return path_join(sys.prefix,'lib','python'+sys.version[:3])

def _locateVerDLL(name):
	return _locate('%s%2d.dll'%(name,_pyvers),[],getWindowsPath())

def chdir(dir):
	'''chdir to dir and return old cwd'''
	cwd = os.path.abspath(os.getcwd())
	os.chdir(dir)
	return cwd

class PythonCopier:
	sources = ['py.ico','pyc.ico','pycon.ico','w9xpopen.exe', 'Lib', 'DLLs']
	nocopy = [	'Lib/lib-tk', 'Lib/test', 'Lib/imputil.py', 'Libs',
				'Lib/lib-old', 'Lib/site-packages', 'Lib/email/test',
				'DLLs/tcl83.dll', 'DLLs/tk83.dll',
				'DLLs/cCopy.pyd', 'Dlls/calldll.pyd', 'DLLs/npstruct.pyd',
				'DLLs/_functional.pyd','DLLs/_pyfribidi.pyd','DLLs/fribidi.dll', 'DLLs/libeay32.dll',
				'DLLs/_tkinter.pyd', 'DLLs/gaxtra.pyd','DLLs/old','Dlls/_grabscreen.pyd',
				'DLLs/m2crypto.dll', 'DLLs/_imagingtk.pyd', 'Doc/XTRA','Doc/Grimoire','Doc/Zope',
				]
	
	_pythonDir = path_dirname(sys.executable)

	def doCopy(self,where, srcDir, pythonDir=None):
		'''make a reduced copy of python'''
		if _verbose: print('Copying python from %s & %s\n  --> %s' % (pythonDir,srcDir,where))
		cwd = chdir(where)
		copyOver(self.sources,pythonDir or self._pythonDir,where,srcDir)

		#clean up unwanted things
		for x in self.nocopy:
			p = path_join(where,x)
			kill(p)
			if _verbose>1: print('	   kill(%s)' % p)
		if not isDir('Doc'): os.mkdir('Doc')
		p = _locateVerDLL('Python')
		copy2(p,'DLLs')
		if _verbose>1: print('	%s --> DLLs' % p)
		chdir(cwd)

def copyPIL(where, tk=0):
	import _imaging, PIL
	src = path_dirname(_imaging.__file__)
	for F in ('_imaging.pyd','libjpeg.dll','libz.dll') + (tk and ('_imagingtk.pyd',) or ()):
		condCopyF(F,src,path_join(where,'Dlls'))
	src = path_join(path_dirname(PIL.__file__),'..')
	copyOver(('PIL',),src,path_join(where,'Lib','site-packages'))

def copy_rl_remove(where):
	from rlextra.utils import _rl_remove
	src = path_dirname(_rl_remove.__file__)
	copyOver(('_rl_remove.py',),src,path_join(where,'Lib'))

def _svnUpdate(svn, url, opts='',target='', rev='',report=None):
	url = string.replace(url,'\\','/')
	if url.endswith('/'): url = url[:-1]
	if target:
		module = target
	else:
		module = url.split('/')[-1]
	if os.path.exists(module):
		act = 'up'
		url = module
	else:
		act = 'co'
	if rev: act += rev
	if opts: act = '%s %s' % (act,opts)
	act = '%s %s %s' % (svn, act, url)
	if target: act = '%s %s' % (act,target)
	do_exec(act,report=report)

def _svnRevSplit(x,pfx='@'):
	x = x or ''
	if '@' in x:
		x, r = x.rsplit('@',1)
		r = pfx+r
	else:
		r = ''
	return x, r

def svnUpdate(svn, url, L, opts='',target=''):
	url, rev = _svnRevSplit(url,' -r ')
	if not L:
		_svnUpdate(svn, url, opts=opts,target=target,rev=rev)
	else:
		for x in L:
			_svnUpdate(svn, '%s/%s' % (url,x), opts=opts, target=target,rev=rev)

def svn_update(workDir, url=None, L=None,opts='',target=''):
	svn = find_exe('svn')
	try:
		os.makedirs(workDir)
	except:
		if not isDir(workDir): raise
	cwd = chdir(workDir)
	try:
		svnUpdate(svn, url, L, opts=opts,target=target)
	finally:
		chdir(cwd)

class SVN:
	exeName = 'svn'
	exe = None
	excludePats=('/**/*.pyc','/**/*.pyo','/**/.svn/**/*')
	def __init__(self, url, target=None, distro=None, copySrc=None, copyTarget=None, fetchOpts='',
						revision=None,excludePats=(),includePats=(),report=None, verbose=False,makePkg=False):
		'''
		url is the repository source URL
		target is an optional staging area subdirectorty to check out in, defaults to last URL component
		distro is an option FileSet filter for the copying, paths should be relative to the copy position'
		copySrc a path to add to target to start the copy from
		copyTarget where in the destination to copy our tree to.
		fetchOpts	additional options for the fetch from repository operation
		revision	what SVN revsion to checkout
		'''

		self.url = url
		self.target = target
		self.distro = distro
		self.copySrc = copySrc
		self.copyTarget = copyTarget
		self.fetchOpts = fetchOpts
		self.revision = revision
		self.excludePats = tuple(set(tuple(excludePats)+self.excludePats))
		self.includePats = tuple(set(tuple(includePats)))
		self.report = report
		self.verbose = verbose
		self.makePkg = makePkg
		if not self.exe:
			self.__class__.exe = find_exe(self.exeName)

	def fetch(self,stagingDir=None):
		'''fetch or update self into dir dst'''
		try:
			os.makedirs(stagingDir)
		except:
			if not isDir(stagingDir): raise
		cwd = chdir(stagingDir)
		try:
			self.update()
		finally:
			chdir(cwd)

	def update(self):
		_svnUpdate(self.exe, self.url, opts=self.fetchOpts, target=self.target, rev=self.revision, report=self.report)

	def copy(self,stagingDir,dstDir,pfb=False):
		FS = self.distro
		report = self.report
		if not os.path.isabs(stagingDir):
			stagingDir = os.path.abspath(stagingDir)
		if not os.path.isabs(dstDir):
			dstDir = os.path.abspath(dstDir)
		target = self.target or self.url.split('/')[-1]
		copySrc = self.copySrc
		if copySrc:
			target = '/'.join((target,copySrc))
			stagingDir = '/'.join((stagingDir,'/'.join(target.split('/')[:-1])))
			stagingDir = stagingDir.replace('/',os.sep)
			target = copySrc.split('/')[-1]
		if not FS or FS=='[skiptop]':
			from rlextra.utils.superglob import FileSet
			fs = FileSet(stagingDir)
			if FS=='[skiptop]':
				fs = FileSet('/'.join((stagingDir,target)))
				target = '.'
			fs.include(target+'/**/*')
			for x in self.excludePats:
				fs.exclude(target+x)
			for x in self.includePats:
				fs.include(target+x)
		else:
			fs = FS(stagingDir)

		if self.copyTarget:
			dstDir = os.path.join(dstDir,self.copyTarget)
		try:
			os.makedirs(dstDir)
		except:
			if not isDir(dstDir): raise
		if self.verbose>2 and report:
			report('####### copying: %s(%r)' % (fs.__class__.__name__,fs.baseDir))
			report('        filenames = %s' % '\n'.join(fs.getFileNames()))
		fs.copyFiles(dstDir)
		if pfb:
			if report: report('##### fetchPFB(%r)'% dstDir)
			fetchPFB(dstDir,report=report)
		makePkg = getattr(self,'makePkg',None)
		if makePkg:
			if not isinstance(makePkg,(tuple,list)):
				makePkg = (makePkg,)
			for pkg in makePkg:
				if not pkg: continue
				if isinstance(pkg,str):
					pkgFn = os.path.join(dstDir,pkg,'__init__.py')
				else:
					pkgFn = os.path.join(dstDir,target.split('/')[-1],'__init__.py')
				if self.verbose>2 and report:
					report('##### make package file %r' % pkgFn)
				if not os.path.isfile(pkgFn):
					open(pkgFn,'w').close()

class HG(SVN):
	exe = None
	exeName = 'hg'
	excludePats = ('/**/*.pyc','/**/*.pyo','/**/.htags','/**/.hgignore','/**/.hg/**/*')
	def __init__(self, url, target=None, distro=None, copySrc=None, copyTarget=None,
					fetchOpts='', revision=None, branch=None,excludePats=(),
					includePats=(),report=None,verbose=False,makePkg=False):
		SVN.__init__(self, url, target=target, distro=distro,
						copySrc=copySrc, copyTarget=copyTarget,
						fetchOpts=fetchOpts, revision=revision,excludePats=excludePats,
						includePats=includePats,report=report,verbose=verbose,makePkg=makePkg)
		self.branch = branch

	def update(self):
		target = self.target
		rev = self.revision
		branch = self.branch
		opts = self.fetchOpts
		url = self.url.replace('\\','/')
		if url.endswith('/'): url = url[:-1]
		if target:
			module = target
		else:
			module = url.split('/')[-1]
		C = [self.exe]
		aC = C.append
		if os.path.exists(module):
			aC('pull -q')
			cmd = ' '.join(C)
			cwd = chdir(module)
			try:
				r = do_exec(cmd,report=self.report)
			finally:
				chdir(cwd)
		else:
			aC('clone')
			aC(url)
			if target: aC(target)
			cmd = ' '.join(C)
			r =  do_exec(cmd,report=self.report)
		if r: return r
		if not branch: branch ='default'
		C = [self.exe,'up',branch]
		aC = C.append
		if rev: aC('-r '+rev)
		if opts: aC(opts)
		cmd = ' '.join(C)
		cwd = chdir(module)
		try:
			return do_exec(cmd,report=self.report)
		finally:
			chdir(cwd)

def svn_info(what,withVersion=True):
	'''return the svn info output for a working file/folder as a dictionary'''
	cwd = os.getcwd()
	if not what: what = cwd
	if isDir(what):
		wd = what
		obj = ''
	elif isFile(what):
		wd = path_dirname(what)
		obj = '"%s"' % path_basename(what)
	else:
		raise ValueError('Not file or folder: "%s"' % what)
	from io import StringIO
	out = StringIO()
	if withVersion: out1 = StringIO()
	try:
		chdir(wd)
		do_exec('%s info %s' % (find_exe('svn'), obj), out)
		if withVersion:
			do_exec('%s %s' % (find_exe('svnversion'), obj or '.'), out1)
	finally:
		chdir(cwd)
	D = {}
	out = out.getvalue().split('\n')
	for o in out:
		o = o.split(':')
		D[o[0].strip()] = ':'.join(o[1:]).strip()
	if withVersion: D['svnversion'] = out1.getvalue().strip() or '?????'
	return D

def _findSVNURL(W):
	'''return the URL corresponding to working folder W'''
	try:
		D = svn_info(W,withVersion=False)
		return D['URL']
	except:
		raise ValueError('Can\'t find SVN URL for working dir '+W)

def _locate_trunk(url, starts=('/branches/','/tags/')):
	from subprocess import Popen, PIPE
	svn = find_exe_nq('svn')
	for s in starts:
		i = url.find(s)
		if i<0: continue
		pfx = url[:i]
		if pfx.endswith('/'): pfx = pfx[:-1]
		SFX = url[i+len(s):].split('/')[1:]
		while SFX:
			u = '/'.join([pfx,'trunk']+SFX)
			p = Popen((svn,'ls',u),stdout=PIPE,stderr=PIPE)
			i = p.wait()
			if not i: return u
			del SFX[0]
	raise ValueError('Cannot locate trunk version of\n\t'+url)

def findSVNURL(W, branch=None):
	url = _findSVNURL(W)
	branch, rev = _svnRevSplit(branch)
	s = '/trunk/'
	i = -1
	if branch:
		if branch.startswith('/'): branch = branch[1:]
		if branch and branch!='trunk':
			if not (branch.startswith('branches/') or branch.startswith('tags/')):
				raise ValueError('branch should start with "branches/" or "tags/" not "%s"' % branch)
			if branch.endswith('/'): branch = branch[:-1]
			i=url.find(branch)
	if i>=0:
		i -= 1
		url = url[:i]+s+url[i+len(branch)+2:]
	else:
		i = url.find(s)
		if i<0:
			url = _locate_trunk(url,('branches','tags'))
			i = url.find(s)
	if branch=='trunk': branch = ''
	if branch:
		url = '%s/%s/%s' % (url[:i],branch,url[i+len(s):])
	if rev: url += rev

	return url

def import_free_call(func,*args,**kw):
	'''eliminate unwanted import side effects and return func()'''
	try:
		OLD = {}
		for x in 'path','modules','meta_path','path_hooks','path_importer_cache':
			try:
				_ = getattr(sys,x)
				if hasattr(_,'copy'): _ = _.copy()
				else: _ = _[:]
				OLD[x] = _
			except:
				pass
		return func(*args,**kw)
	finally:
		for x,v in list(OLD.items()):
			_ = getattr(sys,x)
			if hasattr(_,'copy'):
				_.clear()
				_.update(v)
			else:
				_[:] = v

def inferSVNURL(URL,modulename):
	'''try to infer the SVN URL from a module import location'''
	def func(URL,modulename):
		try:
			NS = {}
			rl_exec('import %s as mod' % modulename,NS)
			mod = NS['mod']
			del NS
			from urllib.parse import urlparse, urlunparse
			host = urlparse(findSVNURL(path_dirname(mod.__file__)),allow_fragments=0)
			url = urlparse(URL)
			return urlunparse((url[0],host[1])+url[2:])
		except:
			return URL
	return import_free_call(func,URL,modulename)

def _fetchRLPrivate(srcDir,branchName,desiredBN=None,what='rlextra',report=None):
		if not desiredBN: desiredBN = branchName
		else:
			if desiredBN.startswith('/'): desiredBN = desiredBN[1:]
			if not (desiredBN=='trunk' or desiredBN.startswith('branches/') or desiredBN.startswith('tags/')):
				raise ValueError('desiredBN should start with "branches/" or "tags/" not "%s"' % desiredBN)
		desiredBN, rev = _svnRevSplit(desiredBN)
		url = inferSVNURL('https://svn.reportlab.com/svn/private/%s/%s' % (desiredBN,what),'rlextra')+rev
		if report: report('Fetching %s from %s'%(what,url))
		svn_update(srcDir,url=url,L=None)
		if report: report('%s fetched' % what)

def fetchPFB(dstDir,report=None):
	#fetch the standard pfb files into ./fonts
	from reportlab.pdfbase._fontdata import _font2fnrMapWin32, findT1File
	N = []
	for fontName in list(_font2fnrMapWin32.keys()):
		try:
			fn = findT1File(fontName)
			if fn is None: raise ValueError
		except:
			N.append(fontName)
			continue
		if sys.platform=='win32':
			fn = path_join(path_dirname(fn),path_basename(fn).lower())
		dd = path_join(dstDir,'reportlab','fonts')
		_src2dst(fn,dd)
		os.chmod(path_join(dd,path_basename(fn)),0o666)
	if N: raise ValueError('Can\'t copy pfb fonts '+ ' '.join(N)) 
	if report: report('pfb font files copied to fonts')

def fetchRL(srcDir,rl=1,rlx=1,rla=None,report=None,pfb=None, branchName=None, branchNameX=None,
				rli=0, branchNameI=None):
	'''convenience routine assumes our reportlab & rlextra (& rlinfra) are SVN controlled'''
	if not branchName: branchName = 'trunk'
	else:
		if branchName.startswith('/'): branchName = branchName[1:]
		if not (branchName=='trunk' or branchName.startswith('branches/') or branchName.startswith('tags/')):
			raise ValueError('branchName should start with "branches/" or "tags/" not "%s"' % branchName)
	if rl or rla:
		branchName, rev = _svnRevSplit(branchName)
		if ('/nomtrail' in branchName or '/isa2009' in branchName or '/isa2008' in branchName
				or '/cmaimp' in branchName or '/isa2010' in branchName or '/nov2010' in branchName
				or '/kiid' in branchName or '/rdr' in branchName or '/cstan' in branchName 
				or '/mstar2012' in branchName):
			url = ('https://svn.reportlab.com/svn/public/reportlab/%s/reportlab' % branchName)+rev
		else:
			url = ('https://svn.reportlab.com/svn/public/reportlab/%s/src/reportlab' % branchName)+rev
		if rl:
			if report: report('Fetching reportlab from '+url)
			svn_update(srcDir,url=url,L=None)
			if report: report('reportlab fetched')
			if pfb:
				fetchPFB(srcDir,report=report)
		if rla:
			if report: report('Fetching rl_addons from '+url)
			url = 'https://svn.reportlab.com/svn/public/reportlab/%s/src/rl_addons' % branchName+rev
			svn_update(srcDir,url=url,L=None)
			if report: report('rl_addons fetched')
	if rlx:
		_fetchRLPrivate(srcDir,branchName,desiredBN=branchNameX,what='rlextra',report=report)
	if rli:
		_fetchRLPrivate(srcDir,branchName,desiredBN=branchNameI,what='rlinfra',report=report)

def stagingArea(suffix=''):
	from reportlab.lib.utils import get_rl_tempdir
	return get_rl_tempdir('SA_'+suffix)

def copyRL(where,srcDir,rl=1,rlx=1,noFetch=0):
	'''copy reportlab or rlextra from the staging area to the result area'''
	if not noFetch: fetchRL(srcDir,rl=rl,rlx=rlx)	#get export to srcDir
	cwd = chdir(where)
	if _verbose: print('Copying RL from %s\n  --> %s' % (srcDir,where))
	copyOver((rl and ('reportlab',) or ())+(rlx and ('rlextra',) or ()), srcDir, where,None)
	chdir(cwd)

def make_stub(stubname,fn,unix=0):
	stubs= {'rml2pdf':'''import sys
from rlextra.rml2pdf import rml2pdf
from reportlab import rl_config
fn=sys.argv[1:]
x = filter(lambda x:x[:9]=='--outdir=',fn)
outDir=None
if x:
	outDir=x[-1][9:]
	map(fn.remove,x)
rml2pdf.main(quiet=rl_config.verbose==0,outDir=outDir,exe=1,fn=fn)
''',
	'pagecatcher': 'from rlextra.pageCatcher.pageCatcher import scriptInterp\nscriptInterp()\n',
	'pythonpoint': 'from reportlab.tools.pythonpoint.pythonpoint import main\nmain()\n',
	}
	open(fn,'w').write((unix and '#!/usr/bin/env python\n' or '') + stubs[stubname])

def postCopyFixupRL(where):
	cwd = chdir(where)
	if sys.hexversion>=0x02020000:
		src = path_join(PythonCopier._pythonDir,'Lib','site-packages')
		for F in ('_renderPM.pyd','_rl_accel.pyd','pyRXP.pyd','sgmlop.pyd','pyHnj.pyd'):
			condCopyF(F,src,'Dlls')
	sys.path.insert(0,where)
	import reportlab.rl_config
	sys.path.remove(where)
	L = []
	for d in dir(reportlab.rl_config):
		if d in ('__builtins__','sys','os','string'): continue
		L.append(d)
	src = path_join(where,'reportlab','rl_config.py')
	dst = path_join(where,'_rl_config.py')
	rename(src,dst)
	open(src,'w').write('from _rl_config import %s\n' % ', '.join(L))
	make_stub('rml2pdf',path_join(where,'rlextra','rml2pdf','exe_stub_rml2pdf_distro.py'))
	make_stub('pagecatcher', path_join(where,'rlextra','pageCatcher','exe_stub_pageCatcher_distro.py'))
	make_stub('pythonpoint',path_join(where,'reportlab','tools','pythonpoint', 'exe_stub_pythonpoint_distro.py'))
	chdir(cwd)

def _compileApp(project,app,zip):
	from rlextra.utils.appmonitor import AppMonitor
	from rlextra.utils.superglob import FileSet
	FS = FileSet(app)
	FS.include('**/*')
	m = AppMonitor(dirname=app,id=project,fileList=FS.getFileNames())
	m._force_ssmName = 0
	m.compileApp()
	m.writeSignedTargetManifest()

	#assembly should now be complete we build everything into an installer
	print('Creating application archive',zip)
	m.makeTargetArchive(zip)

def _compileAppX(ifn,project,app,zip):
	'''special purpose function designed to be run externally'''
	from io import StringIO
	sys.stdout = StringIO()
	sys.stderr = StringIO()
	try:
		_compileApp(project,app,zip)
	except:
		traceback.print_exc()
	from marshal import dump
	f = open(ifn,'wb')
	dump((sys.stdout.getvalue(),sys.stderr.getvalue()),f)
	f.close()

def _externalRun(ifn, cmd, PPATH, verbose=1, errMsg=None):
	if PPATH: os.environ['PYTHONPATH'] = PPATH
	try:
		f = os.popen(cmd,'w')
		#status= \
		f.close()
	finally:
		if PPATH: del os.environ['PYTHONPATH']
	f = open(ifn,'rb')
	from marshal import load
	out,err = load(f)
	f.close()
	if verbose:
		print(out)
		print(err)
	if err and errMsg:
		raise ValueError(errMsg+('\n%s'%cmd))
	return out,err

def makePat(P):
	'''
	P is a list of	glob style patterns
	return the corresponding compiled re
	'''
	import re, fnmatch
	L=[]
	for p in P:
		L.append(fnmatch.translate(p))
	return re.compile('^(?:%s)$'%string.join(L,'|'))

def makeRLDocs(where):
	cwd = chdir(path_join(where,'reportlab','docs'))
	opp = os.environ.get('PYTHONPATH',None)
	pp = where
	if opp: pp += ';'+opp
	os.environ['PYTHONPATH'] = pp
	do_exec(sys.executable + ' genAll.py')
	PDF = glob.glob('*.pdf')
	if _verbose>2: print('pdf=',PDF)
	if opp: os.environ['PYTHONPATH'] = opp
	else: del os.environ['PYTHONPATH']
	chdir(cwd)
	chdir(path_join(where,'rlextra','graphics','doc'))
	do_exec('%s %s diagradoc.rml' % (sys.executable, path_join(where,'rlextra','rml2pdf','rml2pdf.py')))
	chdir(cwd)
	return PDF

_binexts = ('dll', 'pyc', 'pyo', 'pyz', 'exe')
def _resource_fixcase(name, path):
	parts = string.split(name, '.')
	name = parts[-1]
	if name in _binexts:
		name = parts[-2]
	d, bnm = os.path.split(path)
	nm, ext = os.path.splitext(bnm)
	if string.lower(name) == string.lower(nm):
		return path_join(d, name + ext)
	return path

def GetCompiled(seq, lvl='c'):
	"""SEQ is a list of .py files, or a logical TOC.
	 Return as .pyc or .pyo files (LVL) after ensuring their existence"""
	if len(seq) == 0: return seq
	rslt = []
	isTOC = type(seq[0])==type(())
	for py in seq:
		if isTOC:
			(nm, fnm), rest = py[:2], py[2:]
		else:
			fnm = py
			nm = os.path.splitext(fnm)[0]
		fnm = os.path.splitext(fnm)[0] + '.py'
		cmpl = 1
		pyc = fnm + lvl
		if os.path.exists(pyc):
			pytm = int(os.stat(fnm)[8])
			ctm = int(os.stat(pyc)[8])
			if pytm < ctm: cmpl = 0
		if cmpl:
			fnm = _resource_fixcase(nm, fnm)
			import py_compile
			py_compile.compile(fnm, pyc)
		if isTOC:
			rslt.append((nm, pyc)+rest)
		else:
			rslt.append(pyc)
	return rslt

def assembleArchive(pkg,toc,fullName=0):
	'''
	pkg is the name of the archive file
	toc is the list of contents path, type
	'''
	from rlextra.utils.installer import CArchive
	A = CArchive()
	L = []
	for p, t in toc:
		n = os.path.splitext(path_basename(p))[0]
		c = {'s':2,'m':0,'b':1,'x':1,'a':0,'z':0}[t]
		if t=='b' or (fullName and t in 'ms'): n = n + os.path.splitext(p)[1]
		if t=='m': L += GetCompiled([(n, p, c, t)])
		else: L.append((n, p, c, t))
	A.build(pkg,L)
	return A

def copyFiles(src,dst,mode='wb'):
	w = open(dst,mode)
	for r in src:
		w.write(open(r,'rb').read())

_MAGIC = {'RLP':"RLP\02\04\03\014\015", 'RLW':"RLW\01\02\03\017\016", 'RLE':"RLE\07\04\01\013\012"}
def _wrapEXE(src,dst,magic,project,script=None,version=None,icon=None):
	v = sys.version
	v = int(v[0])*10+int(v[2])
	if _verbose: print('making %s(%d) from %s icon=%s version=%s' % (dst,v,src,icon,version))
	d = open(dst,'wb')
	d.write(open(src,'rb').read())
	if script:
		d.write(script)
		s = struct.pack('<ihh',len(script),len(project),v)
	else:
		s = struct.pack('<hh',len(project),v)
	d.write(project)
	d.write(s)
	d.write(magic)
	d.close()
	if icon:
		odst = dst+'.tmp'
		os.rename(dst,odst)
		import pywintypes
		try:
			from rlextra.utils.installer.icon import CopyIcons
			CopyIcons(dst, icon)
		except ImportError:
			raise ValueError("win32api & NT/W2K are required for updating icons.")
		except pywintypes.error:
			raise ValueError("NT/Win2k required to update icons")
		os.remove(odst)
	if version:
		odst = dst+'.tmp'
		os.rename(dst,odst)
		import pywintypes
		try:
			from rlextra.utils.installer.versionInfo import SetVersion
			SetVersion(dst, version)
		except ImportError:
			raise ValueError("win32api & NT/W2K are required for setting Version resources.")
		except pywintypes.error:
			raise ValueError("NT/Win2k required to update version resource")
		os.remove(odst)

def _makeEXE(where,n,debug,magic,project,script=None,version=None,icon=None,stub=None,ext='exe'):
	supDir = installerDir('support')
	if ext and ext[0]!='.': ext = '.'+ext
	w = (stub or n)+(debug and '_d' or '')+ext
	n = n+ext
	src = path_join(supDir,w)
	dst = path_join(where,n)
	_wrapEXE(src,dst,magic,project,script=script,version=version,icon=icon)

def makeEXES(where,PROJECT,RLP=('python','pythonw',),RLW=(),debug=1):
	if _verbose: print('Making exe files in ',where)
	for n,p,w,v,i in RLW:
		_makeEXE(where,n,debug,
				_MAGIC['RLW'],
				PROJECT,
				script=open(p,'r').read(),
				icon=i,
				version=v,
				stub=w and 'rlwrapw' or 'rlwrap')
	for n in RLP:
		_makeEXE(where,n,debug,_MAGIC['RLP'],PROJECT)

def getHere():
	global _cached_getHere
	if '_cached_getHere' not in list(globals().keys()):
		_cached_getHere = getFlag('--where',1,sys.argv[0])
		if not os.path.isabs(_cached_getHere): _cached_getHere = os.path.abspath(_cached_getHere)
		if not isDir(_cached_getHere): _cached_getHere = path_dirname(_cached_getHere)
	return _cached_getHere

def listAllFiles(dir,rel=1):
	def record(F,dir,names):
		F.extend(list(filter(isFile,list(map(lambda x, dir=dir: path_join(dir,x),names)))))
	F = []
	os.path.walk(dir,record,F)
	if rel:
		n = len(dir)+len(os.sep)
		F = list(map(lambda x,n=n: x[n:],F))
	return F

def listAllDirs(dir,rel=1):
	def record(F,dir,names):
		F.extend(list(filter(isDir,list(map(lambda x, dir=dir: path_join(dir,x),names)))))
	F = []
	os.path.walk(dir,record,F)
	if rel:
		n = len(dir)+len(os.sep)
		F = list(map(lambda x,n=n: x[n:],F))
	return F

def cleanEmpties(dir):
	D = listAllDirs(dir)
	D.reverse()
	cwd = chdir(dir)
	for d in D:
		if not os.listdir(d): os.rmdir(d)
	chdir(cwd)

def _forceDir(d):
	D = path_dirname(d)
	if not isDir(D): os.makedirs(D)

def _copyF(s,d):
	if isFile(d):
		remove(d)
	else:
		_forceDir(d)
	copy2(s,d)

def sitecustomize(where):
	'''for versions < 2.2 put site-packages into the path'''
	if sys.hexversion<0x02020000:
		open(path_join(where,'Lib','sitecustomize.py'),'w').write(
'''import sys, os
from site import addsitedir
addsitedir(path_join(os.path.dirname(__file__),'site-packages'))
'''
		)

def copyF(F,src,dst):
	_copyF(os.path.abspath(path_join(src,F)), os.path.abspath(path_join(dst,F)))

def condCopyF(F,src,dst):
	srcF = os.path.abspath(path_join(src,F))
	if isFile(srcF):
		_copyF(srcF, os.path.abspath(path_join(dst,F)))

class PyComCopier:
	sources=(
		'win32/win32api.pyd', 'win32/win32trace.pyd',
		'win32/lib/commctrl.py', 'win32/lib/win32con.py', 'win32/lib/win32traceutil.py', 'win32/lib/winerror.py',
		'win32com/__init__.py', 'win32com/client/CLSIDToClass.py', 'win32com/client/__init__.py', 'win32com/client/build.py',
		'win32com/client/connect.py', 'win32com/client/dynamic.py', 'win32com/client/gencache.py',
		'win32com/client/genpy.py', 'win32com/client/makepy.py', 'win32com/client/selecttlb.py',
		'win32com/client/util.py', 'win32com/olectl.py', 'win32com/server/__init__.py', 'win32com/server/connect.py',
		'win32com/server/dispatcher.py', 'win32com/server/exception.py', 'win32com/server/factory.py', 'win32com/server/localserver.py',
		'win32com/server/policy.py', 'win32com/server/register.py', 'win32com/server/util.py', 'win32com/storagecon.py',
		'win32com/universal.py', 'win32com/util.py',

		# these are only used by gui type things
		#'Pythonwin/pywin/__init__.py', 'Pythonwin/pywin/dialogs/__init__.py', 'Pythonwin/pywin/dialogs/list.py', 'Pythonwin/pywin/dialogs/status.py',
		#'Pythonwin/pywin/mfc/__init__.py', 'Pythonwin/pywin/mfc/dialog.py', 'Pythonwin/pywin/mfc/object.py', 'Pythonwin/pywin/mfc/window.py',
		#'win32com/client/combrowse.py', 'Pythonwin/win32ui.pyd', 'Pythonwin/win32uiole.pyd', 'win32com/client/tlbrowse.py',
		)

	def doCopy(self,where):
		import win32com
		src = os.path.abspath(path_join(path_dirname(win32com.__file__),'..'))
		sp = os.path.abspath(path_join(where,'Lib','site-packages'))
		d = path_join(sp,'win32com.pth')
		_forceDir(d)
		open(d,'w').write('win32\nwin32\\lib\n')
		list(map(lambda x,dst=sp,src=src: copyF(x,src,dst),self.sources))
		for f in ('PythonCOM', 'PyWinTypes'):
			s = _locateVerDLL(f)
			d = path_join(where,'Dlls',string.lower(path_basename(s)))
			_copyF(s,d)
		sitecustomize(where)

def findModules(d,wlen,ipfx='',excl_re=None,incl_list=None):
	from py_compile import compile
	cwd = chdir(d)
	here = os.getcwd()
	R = []
	for m in glob.glob('*.py'):
		n = ipfx+m
		if incl_list and n not in incl_list: continue
		n = n[:-3]
		c = normPath(path_join(here,m + 'c'))
		co = '_p:'+c[wlen:-1]
		if n[-9:]=='.__init__': n = n[:-9]
		if excl_re and excl_re.match(n): continue
		if _verbose>9: print('Adding %s to pyz' % c[:-1])
		compile(m,c,co)
		R.append((n,c))
	P = list(filter(lambda x, d=d: isDir(x) and isFile(path_join(x,'__init__.py')), os.listdir(here)))
	for p in P:
		R.extend(findModules(p,wlen,ipfx+p+'.',excl_re,incl_list))
	chdir(cwd)
	return R

def rename(src,dst):
	remove(dst)
	try:
		os.rename(src,dst)
	except:
		pass

def move_files(root,files):
	for f, d in list(files.items()):
		rename(path_join(root,f),path_join(root,d,path_basename(f)))

def cleanupBeforePYZ(where,movethese={},copythese={},remove_before_pyz=()):
	if _verbose: print('Moving RL files to', where)
	#move to root dir
	move_files(where,movethese)

	for f, d in list(copythese.items()):
		copy2(path_join(where,f),path_join(where,d))

	if _verbose: print('Removing unwanted files')
	list(map(kill,remove_before_pyz))

def makePYZ(where, pyz_exclude=(), remove_after_pyz=(), pyzName=None, preLoad=1, addLib=1, cleanUp=1, incl_list=None, verbose=None):
	if verbose is None: verbose = _verbose
	if verbose: print('Locating modules for PYZ')
	from rlextra.utils.installer._rl_archive import ZlibArchive
	from py_compile import compile
	instDir = installerDir()
	Z = []
	if preLoad:
		for n in ('_rl_archive','_rl_iu','_rl_boot')[:max(2,preLoad)]:
			f = path_join(instDir,n+'.py')
			c = f+'c'
			compile(f,c,f)
			Z.append((n,c))

	excl_re = makePat(pyz_exclude)
	where = os.path.abspath(where)
	wlen = len(where)+len(os.sep)
	lib = path_join(where,'Lib')
	for D in (where,)+(addLib and (lib,) or ()):
		Z.extend(findModules(D,wlen,excl_re=excl_re,incl_list=incl_list))
	if verbose: print('Building PYZ, number of modules=%d' % len(Z))
	ZlibArchive().build(pyzName or path_join(lib,'reportlab.pyz'),Z)
	if cleanUp:
		for m,p in Z[preLoad and max(preLoad,2) or 0:]: #first two belong to us
			remove(p)		#x.pyc
			remove(p[:-1])	#x.py
		list(map(kill,remove_after_pyz))

def compile_files(F,removePy=0):
	'''compile_files(F)
	python compile files in F'''
	from py_compile import compile
	for fn in F:
		compile(fn)
		if removePy: remove(fn)
	return len(F)

def removePYC(F,removePy=0):
	'''removePYC(F)
	remove pyc files parallel to files in F'''
	for f in F:
		f = f.split('/')
		if sys.platform=='mac': f.insert(0,'')
		remove(os.path.splitext(os.sep.join(f))[0]+'.pyc')

def compile_dir(dir, removePy=0, noCompile=[]):
	cwd = os.getcwd()
	if not os.path.isabs(dir): dir = path_join(cwd,dir)
	chdir(dir)
	from rlextra.utils.superglob import FileSet
	fsPy = FileSet(dir)
	fsPy.include('**/*.py')
	fsPy.scan()
	count = 0
	F = fsPy.getFileNames()
	count = compile_files(list(filter(lambda x,NC=noCompile: x not in NC,F)), removePy=removePy)
	removePYC(noCompile)
	if removePy:
		for f in [x for x in F if x[-4:]=='.pyc']:
			remove(f[:-1])
	chdir(cwd)
	return count

def runInno(fileName,where=None):
	Inno = None
	for c in (	'C:\\Program Files\\Inno Setup 5\\Compil32.exe',
				'C:\\Program Files\\Inno Setup 4\\Compil32.exe',
				'C:\\Program Files\\Inno\\Compil32.exe',
				'C:\\Program Files\\My Inno Setup Extensions\\Compil32.exe',
				'C:\\Inno\\Compil32.exe',
				'C:\\My Inno Setup Extensions\\Compil32.exe'):
		if isFile(c):
			Inno = c
			break
	if Inno:
		cmd = [Inno,'/cc',where and path_join(where,fileName) or fileName]
		from distutils.spawn import spawn
		#rc = \
		spawn(cmd,search_path=0,verbose=_verbose,dry_run=0)
	else:
		print('Cannot find Inno, run manually')

def makeTGZ(tgzName, archDir,AFL=(),tarName=None):
	#AR the tar module I found is Python 2.2 only and chokes
	#when gzipping; had to handle that myself
	AFL = AFL or listAllFiles(archDir,rel=0)
	from rlextra.utils import tarfile
	if not tarName:
		from reportlab.lib.utils import get_rl_tempfile
		tarName = get_rl_tempfile()
	arch = tarfile.TarFileCompat(tarName, 'w', compression=tarfile.TAR_PLAIN)
	for fn in AFL: arch.write(fn)
	arch.close()
	import gzip
	if tgzName[:-4]!='.tgz': tgzName = tgzName + '.tgz'
	remove(tgzName)
	g = gzip.GzipFile(tgzName, 'wb', 9)
	f = open(tarName, 'rb')
	while 1:
		chunk = f.read(1024)
		if not chunk: break
		g.write(chunk)
	g.close()
	f.close()
	remove(tarName)

def expandTGZ(tgzName,tarName=None):
	'''Expands tgz archives
	Warning this only handles regular files'''
	import tarfile, gzip
	if not tarName:
		from reportlab.lib.utils import get_rl_tempfile
		tarName = get_rl_tempfile()
	g = gzip.GzipFile(tgzName)
	t = open(tarName, 'wb')
	while 1:
		data = g.read(1024)
		if not data: break
		t.write(data)
	g.close()
	t.close()
	t = tarfile.TarFile(tarName)
	m = t.getmembers()

	for d in filter(lambda x, t=tarfile.REGTYPE: x.type==t, m):
		n = d.name.split('/')
		if n[0]=='.': del n[0]
		if len(n)>1:
			dn = os.sep.join(n[:-1])
			if not isDir(dn): os.makedirs(dn)
		open(os.sep.join(n),'wb').write(t.readstr(d))

	t.close()
	remove(tarName)

def makeZIP(zipName, archDir, AFL=()):
	AFL = AFL or listAllFiles(archDir,rel=0)
	import zipfile
	if zipName[-4:]!='.zip': zipName = zipName + '.zip'
	remove(zipName)
	arch = zipfile.ZipFile(zipName, 'w', compression=zipfile.ZIP_DEFLATED)
	for fn in AFL: arch.write(fn)
	arch.close()

def makeArchive(name, where, what='.', convertCRLF=sys.platform=='win32',zipmode='tgz'):
	"Archives file/dir in 'what', or current directory if not specified"
	cwd = chdir(where)
	if name[-4:]=='.tgz' or name[-4:]=='.tar': name = name[:-4]
	name = path_join(os.getcwd(),name)
	if zipmode=='tgz':
		remove(name+'.tar')
		remove(name+'.tgz')
		tar = find_exe('tar')
		do_exec(tar + ' cf %s.tar %s' % (name,what))
		gzip = find_exe('gzip')
		do_exec(gzip + ' %s.tar' % name)
		rename(name+'.tar.gz', name+'.tgz')
	elif zipmode=='zip':
		zip = find_exe('zip')
		remove(name+'.zip')
		do_exec(zip+' -ur %s.zip . -i %s' % (name,what))
	else:
		raise ValueError('Illegal zipmode %s' % zipmode)
	chdir(cwd)

def generateProductInfo(where):
	cwd = chdir(where)
	python = sys.executable
	for (cmd,dir) in [
		("%(python)s genuserguide.py --outdir=%(where)s/docs", "%(where)s/reportlab/docs/userguide"),
		("%(python)s gen_rmluserguide.py --outdir=%(where)s/docs", "%(where)s/rlextra/rml2pdf/doc"),
		("%(python)s %(where)s/rlextra/rml2pdf/rml2pdf.py %(where)s/rlextra/pageCatcher/PageCatchIntro.rml", "%(where)s/rlextra/pageCatcher"),
		("%(python)s %(where)s/reportlab/tools/pythonpoint/pythonpoint.py %(where)s/reportlab/tools/pythonpoint/demos/pythonpoint.xml", "%(where)s/reportlab/tools/pythonpoint/demos"),
		("%(python)s %(where)s/rlextra/rml2pdf/rml2pdf.py diagradoc.rml", "%(where)s/rlextra/graphics/doc"),
		]:
		cmd = cmd.replace('/',os_sep)
		dir = dir.replace('/',os_sep)
		dir = dir % locals()
		if not isDir(dir): continue
		chdir(dir)
		cmd = cmd % locals()
		if _verbose>5: print('################ %s: start %s' % (time.asctime(time.gmtime()), cmd))
		i, o = os.popen4(cmd,'t')
		t = o.read()
		if _verbose>1 and t: print(t)
		if _verbose>5: print('################ %s: finished' % time.asctime(time.gmtime()))
	chdir(cwd)

_re_exts2del=re.compile(r'.*\.(?:bat|c|cc|cpp|h|py|pyc|pyo|prep)$',re.I)
def cleanProductInfo(dir1, dir2, keep):
	'''
	removes code like files eg, .py, .pyc, .c, .h, etc from underneath dir1
	removes corresponding survivors from under dir2
	'''
	dir1 = os.path.abspath(dir1)
	dir2 = os.path.abspath(dir2)


	K = listAllFiles(dir1)
	D = list(filter(_re_exts2del.match,K))
	for f in D:
		if f in keep:
			D.remove(f)
		else:
			K.remove(f)
	if _verbose>9:
		import pprint
		print('dir1=',dir1, '\nK=')
		pprint.pprint(K)
		print('dir2=',dir2, '\nD=')
		pprint.pprint(D)
	cwd = chdir(dir1)
	list(map(kill,D))
	chdir(dir2)
	list(map(kill,K))
	chdir(cwd)

def getFileActionList(dir,patterns={},L=None,fal=None):
	cwd = chdir(dir)
	if L is None: L = []
	A = [x[1] for x in L]
	for k,p in list(patterns.items()):
		F = list(filter(lambda x,A=A: x not in A,filesMatch(p)))
		A.extend(F)
		for f in F:
			L.append((k,f))
	F = listAllFiles(dir)
	F.sort()
	for f in F:
		fs = f.replace('\\','/')
		if fs in A: continue
		d = path_dirname(f)
		fn, ext = os.path.splitext(f)
		bn = path_basename(fn)
		ext = ext.lower()
		if ext == '.pyc': continue
		if ext=='.py':
			if isFile(path_join(dir,d,'__init__.py')):
				k = 'z'
			else:
				k = 'k' #keep
		elif ext in ('.bat','.exe'):
			k = 'd'
		elif ext=='.pdf':
			if d!='docs':
				k = 'm', path_join(dir,'docs')
			else:
				k = 'k'
		elif bn.lower() in ('readme', 'changes', '00readme'):
			k = 'k'
		elif ext in ('.txt','.rml'):
			k = 'k'
		else:
			k = 'k'
		L.append(type(k) is type('') and (k,fs) or (k[0],fs,k[1].replace('\\','/')))
	L.sort()
	chdir(cwd)
	return L

def writeFileActionList(dir,patterns={},filename='new_fileactions.txt',L=None):
	fa = open(filename,'w')
	for x in getFileActionList(dir,patterns=patterns,L=L):
		print(repr(x)[1:-1], file=fa)

def readFileActionList(filename='fileactions.txt'):
	return list(map(eval,open(filename,'r').readlines()))
